from authlib.consts import default_json_headers
from authlib.jose import JoseError

from ..rfc6749 import AccessDeniedError
from ..rfc6749 import InvalidClientError
from ..rfc6749 import InvalidRequestError
from ..rfc6749 import UnauthorizedClientError
from ..rfc7591 import InvalidClientMetadataError
from ..rfc7591.claims import ClientMetadataClaims


class ClientConfigurationEndpoint:
    ENDPOINT_NAME = "client_configuration"

    def __init__(self, server=None, claims_classes=None):
        self.server = server
        self.claims_classes = claims_classes or [ClientMetadataClaims]

    def __call__(self, request):
        return self.create_configuration_response(request)

    def create_configuration_response(self, request):
        # This request is authenticated by the registration access token issued
        # to the client.
        token = self.authenticate_token(request)
        if not token:
            raise AccessDeniedError()

        request.credential = token

        client = self.authenticate_client(request)
        if not client:
            # If the client does not exist on this server, the server MUST respond
            # with HTTP 401 Unauthorized and the registration access token used to
            # make this request SHOULD be immediately revoked.
            self.revoke_access_token(request, token)
            raise InvalidClientError(
                status_code=401, description="The client does not exist on this server."
            )

        if not self.check_permission(client, request):
            # If the client does not have permission to read its record, the server
            # MUST return an HTTP 403 Forbidden.
            raise UnauthorizedClientError(
                status_code=403,
                description="The client does not have permission to read its record.",
            )

        request.client = client

        if request.method == "GET":
            return self.create_read_client_response(client, request)
        elif request.method == "DELETE":
            return self.create_delete_client_response(client, request)
        elif request.method == "PUT":
            return self.create_update_client_response(client, request)

    def create_endpoint_request(self, request):
        return self.server.create_json_request(request)

    def create_read_client_response(self, client, request):
        body = self.introspect_client(client)
        body.update(self.generate_client_registration_info(client, request))
        return 200, body, default_json_headers

    def create_delete_client_response(self, client, request):
        self.delete_client(client, request)
        headers = [
            ("Cache-Control", "no-store"),
            ("Pragma", "no-cache"),
        ]
        return 204, "", headers

    def create_update_client_response(self, client, request):
        # The updated client metadata fields request MUST NOT include the
        # 'registration_access_token', 'registration_client_uri',
        # 'client_secret_expires_at', or 'client_id_issued_at' fields
        must_not_include = (
            "registration_access_token",
            "registration_client_uri",
            "client_secret_expires_at",
            "client_id_issued_at",
        )
        for k in must_not_include:
            if k in request.payload.data:
                raise InvalidRequestError()

        # The client MUST include its 'client_id' field in the request
        client_id = request.payload.data.get("client_id")
        if not client_id:
            raise InvalidRequestError()
        if client_id != client.get_client_id():
            raise InvalidRequestError()

        # If the client includes the 'client_secret' field in the request,
        # the value of this field MUST match the currently issued client
        # secret for that client.
        if "client_secret" in request.payload.data:
            if not client.check_client_secret(request.payload.data["client_secret"]):
                raise InvalidRequestError()

        client_metadata = self.extract_client_metadata(request)
        client = self.update_client(client, client_metadata, request)
        return self.create_read_client_response(client, request)

    def extract_client_metadata(self, request):
        json_data = request.payload.data.copy()
        client_metadata = {}
        server_metadata = self.get_server_metadata()
        for claims_class in self.claims_classes:
            options = (
                claims_class.get_claims_options(server_metadata)
                if hasattr(claims_class, "get_claims_options") and server_metadata
                else {}
            )
            claims = claims_class(json_data, {}, options, server_metadata)
            try:
                claims.validate()
            except JoseError as error:
                raise InvalidClientMetadataError(error.description) from error

            client_metadata.update(**claims.get_registered_claims())
        return client_metadata

    def introspect_client(self, client):
        return {**client.client_info, **client.client_metadata}

    def generate_client_registration_info(self, client, request):
        """Generate ```registration_client_uri`` and ``registration_access_token``
        for RFC7592. By default this method returns the values sent in the current
        request. Developers MUST rewrite this method to return different registration
        information.::

            def generate_client_registration_info(self, client, request):{
                access_token = request.headers['Authorization'].split(' ')[1]
                return {
                    'registration_client_uri': request.uri,
                    'registration_access_token': access_token,
                }

        :param client: the instance of OAuth client
        :param request: formatted request instance
        """
        raise NotImplementedError()

    def authenticate_token(self, request):
        """Authenticate current credential who is requesting to register a client.
        Developers MUST implement this method in subclass::

            def authenticate_token(self, request):
                auth = request.headers.get("Authorization")
                return get_token_by_auth(auth)

        :return: token instance
        """
        raise NotImplementedError()

    def authenticate_client(self, request):
        """Read a client from the request payload.
        Developers MUST implement this method in subclass::

            def authenticate_client(self, request):
                client_id = request.payload.data.get("client_id")
                return Client.get(client_id=client_id)

        :return: client instance
        """
        raise NotImplementedError()

    def revoke_access_token(self, token, request):
        """Revoke a token access in case an invalid client has been requested.
        Developers MUST implement this method in subclass::

            def revoke_access_token(self, token, request):
                token.revoked = True
                token.save()

        """
        raise NotImplementedError()

    def check_permission(self, client, request):
        """Checks whether the current client is allowed to be accessed, edited
        or deleted. Developers MUST implement it in subclass, e.g.::

            def check_permission(self, client, request):
                return client.editable

        :return: boolean
        """
        raise NotImplementedError()

    def delete_client(self, client, request):
        """Delete authorization code from database or cache. Developers MUST
        implement it in subclass, e.g.::

            def delete_client(self, client, request):
                client.delete()

        :param client: the instance of OAuth client
        :param request: formatted request instance
        """
        raise NotImplementedError()

    def update_client(self, client, client_metadata, request):
        """Update the client in the database. Developers MUST implement this method
        in subclass::

            def update_client(self, client, client_metadata, request):
                client.set_client_metadata(
                    {**client.client_metadata, **client_metadata}
                )
                client.save()
                return client

        :param client: the instance of OAuth client
        :param client_metadata: a dict of the client claims to update
        :param request: formatted request instance
        :return: client instance
        """
        raise NotImplementedError()

    def get_server_metadata(self):
        """Return server metadata which includes supported grant types,
        response types and etc.
        """
        raise NotImplementedError()



from collections import defaultdict

from django.http import HttpRequest
from django.utils.functional import cached_property

from authlib.common.encoding import json_loads
from authlib.oauth2.rfc6749 import JsonPayload
from authlib.oauth2.rfc6749 import JsonRequest
from authlib.oauth2.rfc6749 import OAuth2Payload
from authlib.oauth2.rfc6749 import OAuth2Request


class DjangoOAuth2Payload(OAuth2Payload):
    def __init__(self, request: HttpRequest):
        self._request = request

    @cached_property
    def data(self):
        data = {}
        data.update(self._request.GET.dict())
        data.update(self._request.POST.dict())
        return data

    @cached_property
    def datalist(self):
        values = defaultdict(list)
        for k in self._request.GET:
            values[k].extend(self._request.GET.getlist(k))
        for k in self._request.POST:
            values[k].extend(self._request.POST.getlist(k))
        return values


class DjangoOAuth2Request(OAuth2Request):
    def __init__(self, request: HttpRequest):
        super().__init__(
            method=request.method,
            uri=request.build_absolute_uri(),
            headers=request.headers,
        )
        self.payload = DjangoOAuth2Payload(request)
        self._request = request

    @property
    def args(self):
        return self._request.GET

    @property
    def form(self):
        return self._request.POST


class DjangoJsonPayload(JsonPayload):
    def __init__(self, request: HttpRequest):
        self._request = request

    @cached_property
    def data(self):
        return json_loads(self._request.body)


class DjangoJsonRequest(JsonRequest):
    def __init__(self, request: HttpRequest):
        super().__init__(request.method, request.build_absolute_uri(), request.headers)
        self.payload = DjangoJsonPayload(request)



"""authlib.oauth2.rfc7523.
~~~~~~~~~~~~~~~~~~~~~~

This module represents a direct implementation of
JSON Web Token (JWT) Profile for OAuth 2.0 Client
Authentication and Authorization Grants.

https://tools.ietf.org/html/rfc7523
"""

from .assertion import client_secret_jwt_sign
from .assertion import private_key_jwt_sign
from .auth import ClientSecretJWT
from .auth import PrivateKeyJWT
from .client import JWTBearerClientAssertion
from .jwt_bearer import JWTBearerGrant
from .token import JWTBearerTokenGenerator
from .validator import JWTBearerToken
from .validator import JWTBearerTokenValidator

__all__ = [
    "JWTBearerGrant",
    "JWTBearerClientAssertion",
    "client_secret_jwt_sign",
    "private_key_jwt_sign",
    "ClientSecretJWT",
    "PrivateKeyJWT",
    "JWTBearerToken",
    "JWTBearerTokenGenerator",
    "JWTBearerTokenValidator",
]



import logging

from django.conf import settings
from django.core.cache import cache
from django.http import HttpResponse

from authlib.common.security import generate_token
from authlib.common.urls import url_encode
from authlib.oauth1 import AuthorizationServer as _AuthorizationServer
from authlib.oauth1 import OAuth1Request
from authlib.oauth1 import TemporaryCredential

from .nonce import exists_nonce_in_cache

log = logging.getLogger(__name__)


class BaseServer(_AuthorizationServer):
    def __init__(self, client_model, token_model, token_generator=None):
        self.client_model = client_model
        self.token_model = token_model

        if token_generator is None:

            def token_generator():
                return {
                    "oauth_token": generate_token(42),
                    "oauth_token_secret": generate_token(48),
                }

        self.token_generator = token_generator
        self._config = getattr(settings, "AUTHLIB_OAUTH1_PROVIDER", {})
        self._nonce_expires_in = self._config.get("nonce_expires_in", 86400)
        methods = self._config.get("signature_methods")
        if methods:
            self.SUPPORTED_SIGNATURE_METHODS = methods

    def get_client_by_id(self, client_id):
        try:
            return self.client_model.objects.get(client_id=client_id)
        except self.client_model.DoesNotExist:
            return None

    def exists_nonce(self, nonce, request):
        return exists_nonce_in_cache(nonce, request, self._nonce_expires_in)

    def create_token_credential(self, request):
        temporary_credential = request.credential
        token = self.token_generator()
        item = self.token_model(
            oauth_token=token["oauth_token"],
            oauth_token_secret=token["oauth_token_secret"],
            user_id=temporary_credential.get_user_id(),
            client_id=temporary_credential.get_client_id(),
        )
        item.save()
        return item

    def check_authorization_request(self, request):
        req = self.create_oauth1_request(request)
        self.validate_authorization_request(req)
        return req

    def create_oauth1_request(self, request):
        if request.method == "POST":
            body = request.POST.dict()
        else:
            body = None
        url = request.build_absolute_uri()
        return OAuth1Request(request.method, url, body, request.headers)

    def handle_response(self, status_code, payload, headers):
        resp = HttpResponse(url_encode(payload), status=status_code)
        for k, v in headers:
            resp[k] = v
        return resp


class CacheAuthorizationServer(BaseServer):
    def __init__(self, client_model, token_model, token_generator=None):
        super().__init__(client_model, token_model, token_generator)
        self._temporary_expires_in = self._config.get(
            "temporary_credential_expires_in", 86400
        )
        self._temporary_credential_key_prefix = self._config.get(
            "temporary_credential_key_prefix", "temporary_credential:"
        )

    def create_temporary_credential(self, request):
        key_prefix = self._temporary_credential_key_prefix
        token = self.token_generator()

        client_id = request.client_id
        redirect_uri = request.redirect_uri
        key = key_prefix + token["oauth_token"]
        token["client_id"] = client_id
        if redirect_uri:
            token["oauth_callback"] = redirect_uri

        cache.set(key, token, timeout=self._temporary_expires_in)
        return TemporaryCredential(token)

    def get_temporary_credential(self, request):
        if not request.token:
            return None

        key_prefix = self._temporary_credential_key_prefix
        key = key_prefix + request.token
        value = cache.get(key)
        if value:
            return TemporaryCredential(value)

    def delete_temporary_credential(self, request):
        if request.token:
            key_prefix = self._temporary_credential_key_prefix
            key = key_prefix + request.token
            cache.delete(key)

    def create_authorization_verifier(self, request):
        key_prefix = self._temporary_credential_key_prefix
        verifier = generate_token(36)
        credential = request.credential
        user = request.user
        key = key_prefix + credential.get_oauth_token()
        credential["oauth_verifier"] = verifier
        credential["user_id"] = user.pk
        cache.set(key, credential, timeout=self._temporary_expires_in)
        return verifier



import json

from werkzeug.wrappers import Request as WSGIRequest
from werkzeug.wrappers import Response as WSGIResponse


class MockDispatch:
    def __init__(self, body=b"", status_code=200, headers=None, assert_func=None):
        if headers is None:
            headers = {}
        if isinstance(body, dict):
            body = json.dumps(body).encode()
            headers["Content-Type"] = "application/json"
        else:
            if isinstance(body, str):
                body = body.encode()
            headers["Content-Type"] = "application/x-www-form-urlencoded"

        self.body = body
        self.status_code = status_code
        self.headers = headers
        self.assert_func = assert_func

    def __call__(self, environ, start_response):
        request = WSGIRequest(environ)

        if self.assert_func:
            self.assert_func(request)

        response = WSGIResponse(
            status=self.status_code,
            response=self.body,
            headers=self.headers,
        )
        return response(environ, start_response)



import json
import time

import pytest
from django.http import JsonResponse
from django.test import override_settings

from authlib.common.encoding import to_unicode
from authlib.common.urls import add_params_to_uri
from authlib.integrations.django_oauth1 import ResourceProtector
from authlib.oauth1.rfc5849 import signature
from tests.util import read_file_path

from .models import Client
from .models import TokenCredential


def create_route():
    require_oauth = ResourceProtector(Client, TokenCredential)

    @require_oauth()
    def handle(request):
        user = request.oauth1_credential.user
        return JsonResponse(dict(username=user.username))

    return handle


@pytest.fixture(autouse=True)
def token(user, client, db):
    token = TokenCredential(
        user_id=user.pk,
        client_id=client.client_id,
        oauth_token="valid-token",
        oauth_token_secret="valid-token-secret",
    )
    token.save()
    yield token
    token.delete()


def test_invalid_request_parameters(factory):
    handle = create_route()
    url = "/user"

    # case 1
    request = factory.get(url)
    resp = handle(request)
    data = json.loads(to_unicode(resp.content))
    assert data["error"] == "missing_required_parameter"
    assert "oauth_consumer_key" in data["error_description"]

    # case 2
    request = factory.get(add_params_to_uri(url, {"oauth_consumer_key": "a"}))
    resp = handle(request)
    data = json.loads(to_unicode(resp.content))
    assert data["error"] == "invalid_client"

    # case 3
    request = factory.get(add_params_to_uri(url, {"oauth_consumer_key": "client"}))
    resp = handle(request)
    data = json.loads(to_unicode(resp.content))
    assert data["error"] == "missing_required_parameter"
    assert "oauth_token" in data["error_description"]

    # case 4
    request = factory.get(
        add_params_to_uri(url, {"oauth_consumer_key": "client", "oauth_token": "a"})
    )
    resp = handle(request)
    data = json.loads(to_unicode(resp.content))
    assert data["error"] == "invalid_token"

    # case 5
    request = factory.get(
        add_params_to_uri(
            url, {"oauth_consumer_key": "client", "oauth_token": "valid-token"}
        )
    )
    resp = handle(request)
    data = json.loads(to_unicode(resp.content))
    assert data["error"] == "missing_required_parameter"
    assert "oauth_timestamp" in data["error_description"]


@override_settings(AUTHLIB_OAUTH1_PROVIDER={"signature_methods": ["PLAINTEXT"]})
def test_plaintext_signature(factory):
    handle = create_route()
    url = "/user"

    # case 1: success
    auth_header = (
        'OAuth oauth_consumer_key="client",'
        'oauth_signature_method="PLAINTEXT",'
        'oauth_token="valid-token",'
        'oauth_signature="secret&valid-token-secret"'
    )
    request = factory.get(url, HTTP_AUTHORIZATION=auth_header)
    resp = handle(request)
    data = json.loads(to_unicode(resp.content))
    assert "username" in data

    # case 2: invalid signature
    auth_header = auth_header.replace("valid-token-secret", "invalid")
    request = factory.get(url, HTTP_AUTHORIZATION=auth_header)
    resp = handle(request)
    data = json.loads(to_unicode(resp.content))
    assert data["error"] == "invalid_signature"


def test_hmac_sha1_signature(factory):
    handle = create_route()
    url = "/user"

    params = [
        ("oauth_consumer_key", "client"),
        ("oauth_token", "valid-token"),
        ("oauth_signature_method", "HMAC-SHA1"),
        ("oauth_timestamp", str(int(time.time()))),
        ("oauth_nonce", "hmac-sha1-nonce"),
    ]
    base_string = signature.construct_base_string(
        "GET", "http://testserver/user", params
    )
    sig = signature.hmac_sha1_signature(base_string, "secret", "valid-token-secret")
    params.append(("oauth_signature", sig))
    auth_param = ",".join([f'{k}="{v}"' for k, v in params])
    auth_header = "OAuth " + auth_param

    # case 1: success
    request = factory.get(url, HTTP_AUTHORIZATION=auth_header)
    resp = handle(request)
    data = json.loads(to_unicode(resp.content))
    assert "username" in data

    # case 2: exists nonce
    request = factory.get(url, HTTP_AUTHORIZATION=auth_header)
    resp = handle(request)
    data = json.loads(to_unicode(resp.content))
    assert data["error"] == "invalid_nonce"


@override_settings(AUTHLIB_OAUTH1_PROVIDER={"signature_methods": ["RSA-SHA1"]})
def test_rsa_sha1_signature(factory):
    handle = create_route()

    url = "/user"

    params = [
        ("oauth_consumer_key", "client"),
        ("oauth_token", "valid-token"),
        ("oauth_signature_method", "RSA-SHA1"),
        ("oauth_timestamp", str(int(time.time()))),
        ("oauth_nonce", "rsa-sha1-nonce"),
    ]
    base_string = signature.construct_base_string(
        "GET", "http://testserver/user", params
    )
    sig = signature.rsa_sha1_signature(base_string, read_file_path("rsa_private.pem"))
    params.append(("oauth_signature", sig))
    auth_param = ",".join([f'{k}="{v}"' for k, v in params])
    auth_header = "OAuth " + auth_param

    request = factory.get(url, HTTP_AUTHORIZATION=auth_header)
    resp = handle(request)
    data = json.loads(to_unicode(resp.content))
    assert "username" in data

    # case: invalid signature
    auth_param = auth_param.replace("rsa-sha1-nonce", "alt-sha1-nonce")
    auth_header = "OAuth " + auth_param
    request = factory.get(url, HTTP_AUTHORIZATION=auth_header)
    resp = handle(request)
    data = json.loads(to_unicode(resp.content))
    assert data["error"] == "invalid_signature"



import pytest
from flask import json

from authlib.oauth2.rfc6749.grants import ClientCredentialsGrant
from authlib.oauth2.rfc7523 import JWTBearerClientAssertion
from authlib.oauth2.rfc7523 import client_secret_jwt_sign
from authlib.oauth2.rfc7523 import private_key_jwt_sign
from tests.util import read_file_path


@pytest.fixture(autouse=True)
def server(server):
    class JWTClientCredentialsGrant(ClientCredentialsGrant):
        TOKEN_ENDPOINT_AUTH_METHODS = [
            JWTBearerClientAssertion.CLIENT_AUTH_METHOD,
        ]

    server.register_grant(JWTClientCredentialsGrant)
    return server


@pytest.fixture(autouse=True)
def client(client, db):
    client.set_client_metadata(
        {
            "scope": "profile",
            "redirect_uris": ["https://client.test/authorized"],
            "grant_types": ["client_credentials"],
            "token_endpoint_auth_method": JWTBearerClientAssertion.CLIENT_AUTH_METHOD,
        }
    )
    db.session.add(client)
    db.session.commit()
    return client


def register_jwt_client_auth(server, validate_jti=True):
    class JWTClientAuth(JWTBearerClientAssertion):
        def validate_jti(self, claims, jti):
            return True

        def resolve_client_public_key(self, client, headers):
            if headers["alg"] == "RS256":
                return read_file_path("jwk_public.json")
            return client.client_secret

    server.register_client_auth_method(
        JWTClientAuth.CLIENT_AUTH_METHOD,
        JWTClientAuth("https://provider.test/oauth/token", validate_jti),
    )


def test_invalid_client(test_client, server):
    register_jwt_client_auth(server)
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "client_credentials",
            "client_assertion_type": JWTBearerClientAssertion.CLIENT_ASSERTION_TYPE,
        },
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_client"


def test_invalid_jwt(test_client, server):
    register_jwt_client_auth(server)

    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "client_credentials",
            "client_assertion_type": JWTBearerClientAssertion.CLIENT_ASSERTION_TYPE,
            "client_assertion": client_secret_jwt_sign(
                client_secret="invalid-secret",
                client_id="client-id",
                token_endpoint="https://provider.test/oauth/token",
            ),
        },
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_client"


def test_not_found_client(test_client, server):
    register_jwt_client_auth(server)

    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "client_credentials",
            "client_assertion_type": JWTBearerClientAssertion.CLIENT_ASSERTION_TYPE,
            "client_assertion": client_secret_jwt_sign(
                client_secret="client-secret",
                client_id="invalid-client",
                token_endpoint="https://provider.test/oauth/token",
            ),
        },
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_client"


def test_not_supported_auth_method(test_client, server, client, db):
    register_jwt_client_auth(server)
    client.set_client_metadata(
        {
            "scope": "profile",
            "redirect_uris": ["https://client.test/authorized"],
            "grant_types": ["client_credentials"],
            "token_endpoint_auth_method": "invalid",
        }
    )
    db.session.add(client)
    db.session.commit()
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "client_credentials",
            "client_assertion_type": JWTBearerClientAssertion.CLIENT_ASSERTION_TYPE,
            "client_assertion": client_secret_jwt_sign(
                client_secret="client-secret",
                client_id="client-id",
                token_endpoint="https://provider.test/oauth/token",
            ),
        },
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_client"


def test_client_secret_jwt(test_client, server):
    register_jwt_client_auth(server)
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "client_credentials",
            "client_assertion_type": JWTBearerClientAssertion.CLIENT_ASSERTION_TYPE,
            "client_assertion": client_secret_jwt_sign(
                client_secret="client-secret",
                client_id="client-id",
                token_endpoint="https://provider.test/oauth/token",
                claims={"jti": "nonce"},
            ),
        },
    )
    resp = json.loads(rv.data)
    assert "access_token" in resp


def test_private_key_jwt(test_client, server):
    register_jwt_client_auth(server)
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "client_credentials",
            "client_assertion_type": JWTBearerClientAssertion.CLIENT_ASSERTION_TYPE,
            "client_assertion": private_key_jwt_sign(
                private_key=read_file_path("jwk_private.json"),
                client_id="client-id",
                token_endpoint="https://provider.test/oauth/token",
            ),
        },
    )
    resp = json.loads(rv.data)
    assert "access_token" in resp


def test_not_validate_jti(test_client, server):
    register_jwt_client_auth(server, validate_jti=False)

    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "client_credentials",
            "client_assertion_type": JWTBearerClientAssertion.CLIENT_ASSERTION_TYPE,
            "client_assertion": client_secret_jwt_sign(
                client_secret="client-secret",
                client_id="client-id",
                token_endpoint="https://provider.test/oauth/token",
            ),
        },
    )
    resp = json.loads(rv.data)
    assert "access_token" in resp



from .client_mixin import OAuth2ClientMixin
from .functions import create_bearer_token_validator
from .functions import create_query_client_func
from .functions import create_query_token_func
from .functions import create_revocation_endpoint
from .functions import create_save_token_func
from .tokens_mixins import OAuth2AuthorizationCodeMixin
from .tokens_mixins import OAuth2TokenMixin

__all__ = [
    "OAuth2ClientMixin",
    "OAuth2AuthorizationCodeMixin",
    "OAuth2TokenMixin",
    "create_query_client_func",
    "create_save_token_func",
    "create_query_token_func",
    "create_revocation_endpoint",
    "create_bearer_token_validator",
]



import time

try:
    import cPickle as pickle
except ImportError:
    import pickle


class SimpleCache:
    """A SimpleCache for testing. Copied from Werkzeug."""

    def __init__(self, threshold=500, default_timeout=300):
        self.default_timeout = default_timeout
        self._cache = {}
        self.clear = self._cache.clear
        self._threshold = threshold

    def _prune(self):
        if len(self._cache) > self._threshold:
            now = time.time()
            toremove = []
            for idx, (key, (expires, _)) in enumerate(self._cache.items()):
                if (expires != 0 and expires <= now) or idx % 3 == 0:
                    toremove.append(key)
            for key in toremove:
                self._cache.pop(key, None)

    def _normalize_timeout(self, timeout):
        if timeout is None:
            timeout = self.default_timeout
        if timeout > 0:
            timeout = time.time() + timeout
        return timeout

    def get(self, key):
        try:
            expires, value = self._cache[key]
            if expires == 0 or expires > time.time():
                return pickle.loads(value)
        except (KeyError, pickle.PickleError):
            return None

    def set(self, key, value, timeout=None):
        expires = self._normalize_timeout(timeout)
        self._prune()
        self._cache[key] = (expires, pickle.dumps(value, pickle.HIGHEST_PROTOCOL))
        return True

    def delete(self, key):
        return self._cache.pop(key, None) is not None

    def has(self, key):
        try:
            expires, value = self._cache[key]
            return expires == 0 or expires > time.time()
        except KeyError:
            return False



import pytest

from authlib.oauth2.rfc6749.grants import ImplicitGrant

authorize_url = "/oauth/authorize?response_type=token&client_id=client-id"


@pytest.fixture(autouse=True)
def server(server):
    server.register_grant(ImplicitGrant)
    return server


@pytest.fixture(autouse=True)
def client(client, db):
    client.set_client_metadata(
        {
            "redirect_uris": ["https://client.test/authorized"],
            "scope": "profile",
            "response_types": ["token"],
            "grant_types": ["implicit"],
            "token_endpoint_auth_method": "none",
        }
    )
    db.session.add(client)
    db.session.commit()
    return client


def test_get_authorize(test_client):
    rv = test_client.get(authorize_url)
    assert rv.data == b"ok"


def test_confidential_client(test_client, db, client):
    client.client_secret = "client-secret"
    client.set_client_metadata(
        {
            "redirect_uris": ["https://client.test/authorized"],
            "scope": "profile",
            "response_types": ["token"],
            "grant_types": ["implicit"],
            "token_endpoint_auth_method": "client_secret_basic",
        }
    )
    db.session.add(client)
    db.session.commit()

    rv = test_client.get(authorize_url)
    assert b"invalid_client" in rv.data


def test_unsupported_client(test_client, db, client):
    client.set_client_metadata(
        {
            "redirect_uris": ["https://client.test/authorized"],
            "scope": "profile",
            "response_types": ["code"],
            "grant_types": ["implicit"],
            "token_endpoint_auth_method": "none",
        }
    )
    db.session.add(client)
    db.session.commit()
    rv = test_client.get(authorize_url)
    assert "unauthorized_client" in rv.location


def test_invalid_authorize(test_client, server):
    rv = test_client.post(authorize_url)
    assert "#error=access_denied" in rv.location

    server.scopes_supported = ["profile"]
    rv = test_client.post(authorize_url + "&scope=invalid")
    assert "#error=invalid_scope" in rv.location


def test_authorize_token(test_client):
    rv = test_client.post(authorize_url, data={"user_id": "1"})
    assert "access_token=" in rv.location

    url = authorize_url + "&state=bar&scope=profile"
    rv = test_client.post(url, data={"user_id": "1"})
    assert "access_token=" in rv.location
    assert "state=bar" in rv.location
    assert "scope=profile" in rv.location


def test_token_generator(test_client, app, server):
    m = "tests.flask.test_oauth2.oauth2_server:token_generator"
    app.config.update({"OAUTH2_ACCESS_TOKEN_GENERATOR": m})
    server.load_config(app.config)
    rv = test_client.post(authorize_url, data={"user_id": "1"})
    assert "access_token=c-implicit.1." in rv.location



import json
import os

import pytest

from authlib.common.urls import url_decode
from authlib.common.urls import urlparse
from authlib.oauth2.rfc6749 import errors
from authlib.oauth2.rfc6749 import grants

from .models import Client
from .models import CodeGrantMixin
from .models import OAuth2Code
from .models import User
from .oauth2_server import create_basic_auth


@pytest.fixture(autouse=True)
def server(server):
    class AuthorizationCodeGrant(CodeGrantMixin, grants.AuthorizationCodeGrant):
        TOKEN_ENDPOINT_AUTH_METHODS = [
            "client_secret_basic",
            "client_secret_post",
            "none",
        ]

        def save_authorization_code(self, code, request):
            auth_code = OAuth2Code(
                code=code,
                client_id=request.client.client_id,
                redirect_uri=request.payload.redirect_uri,
                response_type=request.payload.response_type,
                scope=request.payload.scope,
                user=request.user,
            )
            auth_code.save()

    server.register_grant(AuthorizationCodeGrant)
    return server


@pytest.fixture(autouse=True)
def client(user):
    client = Client(
        user_id=user.pk,
        client_id="client-id",
        client_secret="client-secret",
        response_type="code",
        grant_type="authorization_code",
        scope="",
        token_endpoint_auth_method="client_secret_basic",
        default_redirect_uri="https://client.test",
    )
    client.save()
    yield client
    client.delete()


def test_get_consent_grant_client(factory, server, client):
    url = "/authorize?response_type=code"
    request = factory.get(url)
    with pytest.raises(errors.InvalidClientError):
        server.get_consent_grant(request)

    url = "/authorize?response_type=code&client_id=invalid-id"
    request = factory.get(url)
    with pytest.raises(errors.InvalidClientError):
        server.get_consent_grant(request)

    client.response_type = ""
    client.save()
    url = "/authorize?response_type=code&client_id=client-id"
    request = factory.get(url)
    with pytest.raises(errors.UnauthorizedClientError):
        server.get_consent_grant(request)

    url = "/authorize?response_type=code&client_id=client-id&scope=profile&state=bar&redirect_uri=https%3A%2F%2Fclient.test&response_type=code"
    request = factory.get(url)
    with pytest.raises(errors.InvalidRequestError):
        server.get_consent_grant(request)


def test_get_consent_grant_redirect_uri(factory, server):
    base_url = "/authorize?response_type=code&client_id=client-id"
    url = base_url + "&redirect_uri=https%3A%2F%2Fa.c"
    request = factory.get(url)
    with pytest.raises(errors.InvalidRequestError):
        server.get_consent_grant(request)

    url = base_url + "&redirect_uri=https%3A%2F%2Fclient.test"
    request = factory.get(url)
    grant = server.get_consent_grant(request)
    assert isinstance(grant, grants.AuthorizationCodeGrant)


def test_get_consent_grant_scope(factory, server):
    server.scopes_supported = ["profile"]
    base_url = "/authorize?response_type=code&client_id=client-id"
    url = base_url + "&scope=invalid"
    request = factory.get(url)
    with pytest.raises(errors.InvalidScopeError):
        server.get_consent_grant(request)


def test_create_authorization_response(factory, server):
    data = {"response_type": "code", "client_id": "client-id"}
    request = factory.post("/authorize", data=data)
    grant = server.get_consent_grant(request)

    resp = server.create_authorization_response(request, grant=grant)
    assert resp.status_code == 302
    assert "error=access_denied" in resp["Location"]

    grant_user = User.objects.get(username="foo")
    resp = server.create_authorization_response(
        request, grant=grant, grant_user=grant_user
    )
    assert resp.status_code == 302
    assert "code=" in resp["Location"]


def test_create_token_response_invalid(factory, server):
    # case: no auth
    request = factory.post("/oauth/token", data={"grant_type": "authorization_code"})
    resp = server.create_token_response(request)
    assert resp.status_code == 401
    data = json.loads(resp.content)
    assert data["error"] == "invalid_client"

    auth_header = create_basic_auth("client-id", "client-secret")

    # case: no code
    request = factory.post(
        "/oauth/token",
        data={"grant_type": "authorization_code"},
        HTTP_AUTHORIZATION=auth_header,
    )
    resp = server.create_token_response(request)
    assert resp.status_code == 400
    data = json.loads(resp.content)
    assert data["error"] == "invalid_request"

    # case: invalid code
    request = factory.post(
        "/oauth/token",
        data={"grant_type": "authorization_code", "code": "invalid"},
        HTTP_AUTHORIZATION=auth_header,
    )
    resp = server.create_token_response(request)
    assert resp.status_code == 400
    data = json.loads(resp.content)
    assert data["error"] == "invalid_grant"


def test_create_token_response_success(factory, server):
    data = get_token_response(factory, server)
    assert "access_token" in data
    assert "refresh_token" not in data


def test_create_token_response_with_refresh_token(factory, server, client, settings):
    settings.AUTHLIB_OAUTH2_PROVIDER["refresh_token_generator"] = True
    server.load_config(settings.AUTHLIB_OAUTH2_PROVIDER)
    client.grant_type = "authorization_code\nrefresh_token"
    client.save()
    data = get_token_response(factory, server)
    assert "access_token" in data
    assert "refresh_token" in data


def test_insecure_transport_error_with_payload_access(factory, server):
    """Test that InsecureTransportError is raised properly without AttributeError
    when accessing request.payload on non-HTTPS requests (issue #795)."""
    del os.environ["AUTHLIB_INSECURE_TRANSPORT"]

    request = factory.get(
        "https://provider.test/authorize?response_type=code&client_id=client-id"
    )

    with pytest.raises(errors.InsecureTransportError):
        server.get_consent_grant(request)


def get_token_response(factory, server):
    data = {"response_type": "code", "client_id": "client-id"}
    request = factory.post("/authorize", data=data)
    grant_user = User.objects.get(username="foo")
    grant = server.get_consent_grant(request)
    resp = server.create_authorization_response(
        request, grant=grant, grant_user=grant_user
    )
    assert resp.status_code == 302

    params = dict(url_decode(urlparse.urlparse(resp["Location"]).query))
    code = params["code"]

    request = factory.post(
        "/oauth/token",
        data={"grant_type": "authorization_code", "code": code},
        HTTP_AUTHORIZATION=create_basic_auth("client-id", "client-secret"),
    )
    resp = server.create_token_response(request)
    assert resp.status_code == 200
    data = json.loads(resp.content)
    return data



import pytest
from flask import json

from authlib.common.urls import url_decode
from authlib.common.urls import urlparse
from authlib.oauth2.rfc6749.grants import (
    AuthorizationCodeGrant as _AuthorizationCodeGrant,
)

from .models import AuthorizationCode
from .models import CodeGrantMixin
from .models import db
from .models import save_authorization_code
from .oauth2_server import create_basic_header

authorize_url = "/oauth/authorize?response_type=code&client_id=client-id"


@pytest.fixture(autouse=True)
def client(client, db):
    client.set_client_metadata(
        {
            "redirect_uris": ["https://client.test"],
            "scope": "profile address",
            "token_endpoint_auth_method": "client_secret_basic",
            "response_types": ["code"],
            "grant_types": ["authorization_code"],
        }
    )
    db.session.add(client)
    db.session.commit()
    return client


class AuthorizationCodeGrant(CodeGrantMixin, _AuthorizationCodeGrant):
    TOKEN_ENDPOINT_AUTH_METHODS = ["client_secret_basic", "client_secret_post", "none"]

    def save_authorization_code(self, code, request):
        return save_authorization_code(code, request)


@pytest.fixture(autouse=True)
def server(server):
    server.register_grant(AuthorizationCodeGrant)
    return server


def test_get_authorize(test_client):
    rv = test_client.get(authorize_url)
    assert rv.data == b"ok"


def test_invalid_client_id(test_client):
    url = "/oauth/authorize?response_type=code"
    rv = test_client.get(url)
    assert b"invalid_client" in rv.data

    url = "/oauth/authorize?response_type=code&client_id=invalid"
    rv = test_client.get(url)
    assert b"invalid_client" in rv.data


def test_invalid_authorize(test_client, server):
    rv = test_client.post(authorize_url)
    assert "error=access_denied" in rv.location

    server.scopes_supported = ["profile"]
    rv = test_client.post(authorize_url + "&scope=invalid&state=foo")
    assert "error=invalid_scope" in rv.location
    assert "state=foo" in rv.location


def test_unauthorized_client(test_client, client, db):
    client.set_client_metadata(
        {
            "redirect_uris": ["https://client.test"],
            "scope": "profile address",
            "token_endpoint_auth_method": "client_secret_basic",
            "response_types": ["token"],
            "grant_types": ["authorization_code"],
        }
    )
    db.session.add(client)
    db.session.commit()

    rv = test_client.get(authorize_url)
    assert "unauthorized_client" in rv.location


def test_invalid_client(test_client):
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "authorization_code",
            "code": "invalid",
            "client_id": "invalid-id",
        },
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_client"

    headers = create_basic_header("code-client", "invalid-secret")
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "authorization_code",
            "code": "invalid",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_client"
    assert resp["error_uri"] == "https://client.test/error#invalid_client"


def test_invalid_code(test_client):
    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "authorization_code",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_request"

    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "authorization_code",
            "code": "invalid",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_grant"

    code = AuthorizationCode(code="no-user", client_id="code-client", user_id=0)
    db.session.add(code)
    db.session.commit()
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "authorization_code",
            "code": "no-user",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_grant"


def test_invalid_redirect_uri(test_client):
    uri = authorize_url + "&redirect_uri=https%3A%2F%2Fa.c"
    rv = test_client.post(uri, data={"user_id": "1"})
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_request"

    uri = authorize_url + "&redirect_uri=https%3A%2F%2Fclient.test"
    rv = test_client.post(uri, data={"user_id": "1"})
    assert "code=" in rv.location

    params = dict(url_decode(urlparse.urlparse(rv.location).query))
    code = params["code"]
    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "authorization_code",
            "code": code,
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_grant"


def test_invalid_grant_type(test_client, client, db):
    client.client_secret = ""
    client.set_client_metadata(
        {
            "redirect_uris": ["https://client.test"],
            "scope": "profile address",
            "token_endpoint_auth_method": "none",
            "response_types": ["code"],
            "grant_types": ["invalid"],
        }
    )
    db.session.add(client)
    db.session.commit()

    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "authorization_code",
            "client_id": "client-id",
            "code": "a",
        },
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "unauthorized_client"


def test_authorize_token_no_refresh_token(app, test_client, client, db, server):
    app.config.update({"OAUTH2_REFRESH_TOKEN_GENERATOR": True})
    server.load_config(app.config)
    client.set_client_metadata(
        {
            "redirect_uris": ["https://client.test"],
            "scope": "profile address",
            "token_endpoint_auth_method": "none",
            "response_types": ["code"],
            "grant_types": ["authorization_code"],
        }
    )
    db.session.add(client)
    db.session.commit()

    rv = test_client.post(authorize_url, data={"user_id": "1"})
    assert "code=" in rv.location

    params = dict(url_decode(urlparse.urlparse(rv.location).query))
    code = params["code"]
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "authorization_code",
            "code": code,
            "client_id": "client-id",
        },
    )
    resp = json.loads(rv.data)
    assert "access_token" in resp
    assert "refresh_token" not in resp


def test_authorize_token_has_refresh_token(app, test_client, client, db, server):
    app.config.update({"OAUTH2_REFRESH_TOKEN_GENERATOR": True})
    server.load_config(app.config)
    client.set_client_metadata(
        {
            "redirect_uris": ["https://client.test"],
            "scope": "profile address",
            "token_endpoint_auth_method": "client_secret_basic",
            "response_types": ["code"],
            "grant_types": ["authorization_code", "refresh_token"],
        }
    )
    db.session.add(client)
    db.session.commit()

    url = authorize_url + "&state=bar"
    rv = test_client.post(url, data={"user_id": "1"})
    assert "code=" in rv.location

    params = dict(url_decode(urlparse.urlparse(rv.location).query))
    assert params["state"] == "bar"

    code = params["code"]
    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "authorization_code",
            "code": code,
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert "access_token" in resp
    assert "refresh_token" in resp


def test_invalid_multiple_request_parameters(test_client):
    url = (
        authorize_url
        + "&scope=profile&state=bar&redirect_uri=https%3A%2F%2Fclient.test&response_type=code"
    )
    rv = test_client.get(url)
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_request"
    assert resp["error_description"] == "Multiple 'response_type' in request."


def test_client_secret_post(app, test_client, client, db, server):
    app.config.update({"OAUTH2_REFRESH_TOKEN_GENERATOR": True})
    server.load_config(app.config)
    client.set_client_metadata(
        {
            "redirect_uris": ["https://client.test"],
            "scope": "profile address",
            "token_endpoint_auth_method": "client_secret_post",
            "response_types": ["code"],
            "grant_types": ["authorization_code", "refresh_token"],
        }
    )
    db.session.add(client)
    db.session.commit()

    url = authorize_url + "&state=bar"
    rv = test_client.post(url, data={"user_id": "1"})
    assert "code=" in rv.location

    params = dict(url_decode(urlparse.urlparse(rv.location).query))
    assert params["state"] == "bar"

    code = params["code"]
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "authorization_code",
            "client_id": "client-id",
            "client_secret": "client-secret",
            "code": code,
        },
    )
    resp = json.loads(rv.data)
    assert "access_token" in resp
    assert "refresh_token" in resp


def test_token_generator(app, test_client, client, server):
    m = "tests.flask.test_oauth2.oauth2_server:token_generator"
    app.config.update({"OAUTH2_ACCESS_TOKEN_GENERATOR": m})
    server.load_config(app.config)
    client.set_client_metadata(
        {
            "redirect_uris": ["https://client.test"],
            "scope": "profile address",
            "token_endpoint_auth_method": "none",
            "response_types": ["code"],
            "grant_types": ["authorization_code"],
        }
    )
    db.session.add(client)
    db.session.commit()

    rv = test_client.post(authorize_url, data={"user_id": "1"})
    assert "code=" in rv.location

    params = dict(url_decode(urlparse.urlparse(rv.location).query))
    code = params["code"]
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "authorization_code",
            "code": code,
            "client_id": "client-id",
        },
    )
    resp = json.loads(rv.data)
    assert "access_token" in resp
    assert "c-authorization_code.1." in resp["access_token"]



import time

import pytest
from flask import json

from authlib.common.urls import add_params_to_uri
from authlib.oauth1.rfc5849 import signature
from tests.util import read_file_path

from .oauth1_server import Client
from .oauth1_server import TokenCredential
from .oauth1_server import User
from .oauth1_server import create_resource_server


@pytest.fixture(autouse=True)
def user(db):
    user = User(username="foo")
    db.session.add(user)
    db.session.commit()
    yield user
    db.session.delete(user)


@pytest.fixture(autouse=True)
def client(db, user):
    client = Client(
        user_id=user.id,
        client_id="client",
        client_secret="secret",
        default_redirect_uri="https://client.test",
    )
    db.session.add(client)
    db.session.commit()
    yield client
    db.session.delete(client)


@pytest.fixture(autouse=True)
def token(db, user, client):
    tok = TokenCredential(
        user_id=user.id,
        client_id=client.client_id,
        oauth_token="valid-token",
        oauth_token_secret="valid-token-secret",
    )
    db.session.add(tok)
    db.session.commit()


@pytest.mark.parametrize("use_cache", [True, False])
def test_invalid_request_parameters(app, test_client, use_cache):
    create_resource_server(app, use_cache, use_cache)
    url = "/user"

    # case 1
    rv = test_client.get(url)
    data = json.loads(rv.data)
    assert data["error"] == "missing_required_parameter"
    assert "oauth_consumer_key" in data["error_description"]

    # case 2
    rv = test_client.get(add_params_to_uri(url, {"oauth_consumer_key": "a"}))
    data = json.loads(rv.data)
    assert data["error"] == "invalid_client"

    # case 3
    rv = test_client.get(add_params_to_uri(url, {"oauth_consumer_key": "client"}))
    data = json.loads(rv.data)
    assert data["error"] == "missing_required_parameter"
    assert "oauth_token" in data["error_description"]

    # case 4
    rv = test_client.get(
        add_params_to_uri(url, {"oauth_consumer_key": "client", "oauth_token": "a"})
    )
    data = json.loads(rv.data)
    assert data["error"] == "invalid_token"

    # case 5
    rv = test_client.get(
        add_params_to_uri(
            url, {"oauth_consumer_key": "client", "oauth_token": "valid-token"}
        )
    )
    data = json.loads(rv.data)
    assert data["error"] == "missing_required_parameter"
    assert "oauth_timestamp" in data["error_description"]


@pytest.mark.parametrize("use_cache", [True, False])
def test_plaintext_signature(app, test_client, use_cache):
    create_resource_server(app, use_cache, use_cache)
    url = "/user"

    # case 1: success
    auth_header = (
        'OAuth oauth_consumer_key="client",'
        'oauth_signature_method="PLAINTEXT",'
        'oauth_token="valid-token",'
        'oauth_signature="secret&valid-token-secret"'
    )
    headers = {"Authorization": auth_header}
    rv = test_client.get(url, headers=headers)
    data = json.loads(rv.data)
    assert "username" in data

    # case 2: invalid signature
    auth_header = auth_header.replace("valid-token-secret", "invalid")
    headers = {"Authorization": auth_header}
    rv = test_client.get(url, headers=headers)
    data = json.loads(rv.data)
    assert data["error"] == "invalid_signature"


@pytest.mark.parametrize("use_cache", [True, False])
def test_hmac_sha1_signature(app, test_client, use_cache):
    create_resource_server(app, use_cache, use_cache)
    url = "/user"

    params = [
        ("oauth_consumer_key", "client"),
        ("oauth_token", "valid-token"),
        ("oauth_signature_method", "HMAC-SHA1"),
        ("oauth_timestamp", str(int(time.time()))),
        ("oauth_nonce", "hmac-sha1-nonce"),
    ]
    base_string = signature.construct_base_string(
        "GET", "http://localhost/user", params
    )
    sig = signature.hmac_sha1_signature(base_string, "secret", "valid-token-secret")
    params.append(("oauth_signature", sig))
    auth_param = ",".join([f'{k}="{v}"' for k, v in params])
    auth_header = "OAuth " + auth_param
    headers = {"Authorization": auth_header}

    # case 1: success
    rv = test_client.get(url, headers=headers)
    data = json.loads(rv.data)
    assert "username" in data

    # case 2: exists nonce
    rv = test_client.get(url, headers=headers)
    data = json.loads(rv.data)
    assert data["error"] == "invalid_nonce"


@pytest.mark.parametrize("use_cache", [True, False])
def test_rsa_sha1_signature(app, test_client, use_cache):
    create_resource_server(app, use_cache, use_cache)
    url = "/user"

    params = [
        ("oauth_consumer_key", "client"),
        ("oauth_token", "valid-token"),
        ("oauth_signature_method", "RSA-SHA1"),
        ("oauth_timestamp", str(int(time.time()))),
        ("oauth_nonce", "rsa-sha1-nonce"),
    ]
    base_string = signature.construct_base_string(
        "GET", "http://localhost/user", params
    )
    sig = signature.rsa_sha1_signature(base_string, read_file_path("rsa_private.pem"))
    params.append(("oauth_signature", sig))
    auth_param = ",".join([f'{k}="{v}"' for k, v in params])
    auth_header = "OAuth " + auth_param
    headers = {"Authorization": auth_header}
    rv = test_client.get(url, headers=headers)
    data = json.loads(rv.data)
    assert "username" in data

    # case: invalid signature
    auth_param = auth_param.replace("rsa-sha1-nonce", "alt-sha1-nonce")
    auth_header = "OAuth " + auth_param
    headers = {"Authorization": auth_header}
    rv = test_client.get(url, headers=headers)
    data = json.loads(rv.data)
    assert data["error"] == "invalid_signature"



from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.asymmetric.ec import SECP256K1
from cryptography.hazmat.primitives.asymmetric.ec import SECP256R1
from cryptography.hazmat.primitives.asymmetric.ec import SECP384R1
from cryptography.hazmat.primitives.asymmetric.ec import SECP521R1
from cryptography.hazmat.primitives.asymmetric.ec import (
    EllipticCurvePrivateKeyWithSerialization,
)
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePrivateNumbers
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePublicKey
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePublicNumbers

from authlib.common.encoding import base64_to_int
from authlib.common.encoding import int_to_base64

from ..rfc7517 import AsymmetricKey


class ECKey(AsymmetricKey):
    """Key class of the ``EC`` key type."""

    kty = "EC"
    DSS_CURVES = {
        "P-256": SECP256R1,
        "P-384": SECP384R1,
        "P-521": SECP521R1,
        # https://tools.ietf.org/html/rfc8812#section-3.1
        "secp256k1": SECP256K1,
    }
    CURVES_DSS = {
        SECP256R1.name: "P-256",
        SECP384R1.name: "P-384",
        SECP521R1.name: "P-521",
        SECP256K1.name: "secp256k1",
    }
    REQUIRED_JSON_FIELDS = ["crv", "x", "y"]

    PUBLIC_KEY_FIELDS = REQUIRED_JSON_FIELDS
    PRIVATE_KEY_FIELDS = ["crv", "d", "x", "y"]

    PUBLIC_KEY_CLS = EllipticCurvePublicKey
    PRIVATE_KEY_CLS = EllipticCurvePrivateKeyWithSerialization
    SSH_PUBLIC_PREFIX = b"ecdsa-sha2-"

    def exchange_shared_key(self, pubkey):
        # # used in ECDHESAlgorithm
        private_key = self.get_private_key()
        if private_key:
            return private_key.exchange(ec.ECDH(), pubkey)
        raise ValueError("Invalid key for exchanging shared key")

    @property
    def curve_key_size(self):
        raw_key = self.get_private_key()
        if not raw_key:
            raw_key = self.public_key
        return raw_key.curve.key_size

    def load_private_key(self):
        curve = self.DSS_CURVES[self._dict_data["crv"]]()
        public_numbers = EllipticCurvePublicNumbers(
            base64_to_int(self._dict_data["x"]),
            base64_to_int(self._dict_data["y"]),
            curve,
        )
        private_numbers = EllipticCurvePrivateNumbers(
            base64_to_int(self.tokens["d"]), public_numbers
        )
        return private_numbers.private_key(default_backend())

    def load_public_key(self):
        curve = self.DSS_CURVES[self._dict_data["crv"]]()
        public_numbers = EllipticCurvePublicNumbers(
            base64_to_int(self._dict_data["x"]),
            base64_to_int(self._dict_data["y"]),
            curve,
        )
        return public_numbers.public_key(default_backend())

    def dumps_private_key(self):
        numbers = self.private_key.private_numbers()
        return {
            "crv": self.CURVES_DSS[self.private_key.curve.name],
            "x": int_to_base64(numbers.public_numbers.x),
            "y": int_to_base64(numbers.public_numbers.y),
            "d": int_to_base64(numbers.private_value),
        }

    def dumps_public_key(self):
        numbers = self.public_key.public_numbers()
        return {
            "crv": self.CURVES_DSS[numbers.curve.name],
            "x": int_to_base64(numbers.x),
            "y": int_to_base64(numbers.y),
        }

    @classmethod
    def generate_key(cls, crv="P-256", options=None, is_private=False) -> "ECKey":
        if crv not in cls.DSS_CURVES:
            raise ValueError(f'Invalid crv value: "{crv}"')
        raw_key = ec.generate_private_key(
            curve=cls.DSS_CURVES[crv](),
            backend=default_backend(),
        )
        if not is_private:
            raw_key = raw_key.public_key()
        return cls.import_key(raw_key, options=options)



from authlib.oauth2 import OAuth2Error


class InteractionRequiredError(OAuth2Error):
    """The Authorization Server requires End-User interaction of some form
    to proceed. This error MAY be returned when the prompt parameter value
    in the Authentication Request is none, but the Authentication Request
    cannot be completed without displaying a user interface for End-User
    interaction.

    http://openid.net/specs/openid-connect-core-1_0.html#AuthError
    """

    error = "interaction_required"


class LoginRequiredError(OAuth2Error):
    """The Authorization Server requires End-User authentication. This error
    MAY be returned when the prompt parameter value in the Authentication
    Request is none, but the Authentication Request cannot be completed
    without displaying a user interface for End-User authentication.

    http://openid.net/specs/openid-connect-core-1_0.html#AuthError
    """

    error = "login_required"


class AccountSelectionRequiredError(OAuth2Error):
    """The End-User is REQUIRED to select a session at the Authorization
    Server. The End-User MAY be authenticated at the Authorization Server
    with different associated accounts, but the End-User did not select a
    session. This error MAY be returned when the prompt parameter value in
    the Authentication Request is none, but the Authentication Request cannot
    be completed without displaying a user interface to prompt for a session
    to use.

    http://openid.net/specs/openid-connect-core-1_0.html#AuthError
    """

    error = "account_selection_required"


class ConsentRequiredError(OAuth2Error):
    """The Authorization Server requires End-User consent. This error MAY be
    returned when the prompt parameter value in the Authentication Request is
    none, but the Authentication Request cannot be completed without
    displaying a user interface for End-User consent.

    http://openid.net/specs/openid-connect-core-1_0.html#AuthError
    """

    error = "consent_required"


class InvalidRequestURIError(OAuth2Error):
    """The request_uri in the Authorization Request returns an error or
    contains invalid data.

    http://openid.net/specs/openid-connect-core-1_0.html#AuthError
    """

    error = "invalid_request_uri"


class InvalidRequestObjectError(OAuth2Error):
    """The request parameter contains an invalid Request Object."""

    error = "invalid_request_object"


class RequestNotSupportedError(OAuth2Error):
    """The OP does not support use of the request parameter."""

    error = "request_not_supported"


class RequestURINotSupportedError(OAuth2Error):
    """The OP does not support use of the request_uri parameter."""

    error = "request_uri_not_supported"


class RegistrationNotSupportedError(OAuth2Error):
    """The OP does not support use of the registration parameter."""

    error = "registration_not_supported"



import pytest

from authlib.common.urls import url_decode
from authlib.common.urls import urlparse
from authlib.oauth2.rfc6749 import errors
from authlib.oauth2.rfc6749 import grants

from .models import Client
from .models import User


@pytest.fixture(autouse=True)
def server(server):
    server.register_grant(grants.ImplicitGrant)
    return server


@pytest.fixture(autouse=True)
def client(user):
    client = Client(
        user_id=user.pk,
        client_id="client-id",
        response_type="token",
        scope="",
        token_endpoint_auth_method="none",
        default_redirect_uri="https://client.test",
    )
    client.save()
    yield client
    client.delete()


def test_get_consent_grant_client(factory, server, client):
    url = "/authorize?response_type=token"
    request = factory.get(url)
    with pytest.raises(errors.InvalidClientError):
        server.get_consent_grant(request)

    url = "/authorize?response_type=token&client_id=invalid-id"
    request = factory.get(url)
    with pytest.raises(errors.InvalidClientError):
        server.get_consent_grant(request)

    client.response_type = ""
    client.save()
    url = "/authorize?response_type=token&client_id=client-id"
    request = factory.get(url)
    with pytest.raises(errors.UnauthorizedClientError):
        server.get_consent_grant(request)


def test_get_consent_grant_scope(factory, server):
    server.scopes_supported = ["profile"]

    base_url = "/authorize?response_type=token&client_id=client-id"
    url = base_url + "&scope=invalid"
    request = factory.get(url)
    with pytest.raises(errors.InvalidScopeError):
        server.get_consent_grant(request)


def test_create_authorization_response(factory, server):
    data = {"response_type": "token", "client_id": "client-id"}
    request = factory.post("/authorize", data=data)
    grant = server.get_consent_grant(request)

    resp = server.create_authorization_response(request, grant=grant)
    assert resp.status_code == 302
    params = dict(url_decode(urlparse.urlparse(resp["Location"]).fragment))
    assert params["error"] == "access_denied"

    grant_user = User.objects.get(username="foo")
    resp = server.create_authorization_response(
        request, grant=grant, grant_user=grant_user
    )
    assert resp.status_code == 302
    params = dict(url_decode(urlparse.urlparse(resp["Location"]).fragment))
    assert "access_token" in params



import pytest

from authlib.oauth2.rfc8414 import AuthorizationServerMetadata
from authlib.oauth2.rfc8414 import get_well_known_url

WELL_KNOWN_URL = "/.well-known/oauth-authorization-server"


def test_well_know_no_suffix_issuer():
    assert get_well_known_url("https://provider.test") == WELL_KNOWN_URL
    assert get_well_known_url("https://provider.test/") == WELL_KNOWN_URL


def test_well_know_with_suffix_issuer():
    assert (
        get_well_known_url("https://provider.test/issuer1")
        == WELL_KNOWN_URL + "/issuer1"
    )
    assert (
        get_well_known_url("https://provider.test/a/b/c") == WELL_KNOWN_URL + "/a/b/c"
    )


def test_well_know_with_external():
    assert (
        get_well_known_url("https://provider.test", external=True)
        == "https://provider.test" + WELL_KNOWN_URL
    )


def test_well_know_with_changed_suffix():
    url = get_well_known_url("https://provider.test", suffix="openid-configuration")
    assert url == "/.well-known/openid-configuration"
    url = get_well_known_url(
        "https://provider.test", external=True, suffix="openid-configuration"
    )
    assert url == "https://provider.test/.well-known/openid-configuration"


def test_validate_issuer():
    #: missing
    metadata = AuthorizationServerMetadata({})
    with pytest.raises(ValueError, match='"issuer" is required'):
        metadata.validate()

    #: https
    metadata = AuthorizationServerMetadata({"issuer": "http://provider.test/"})
    with pytest.raises(ValueError, match="https"):
        metadata.validate_issuer()

    #: query
    metadata = AuthorizationServerMetadata({"issuer": "https://provider.test/?a=b"})
    with pytest.raises(ValueError, match="query"):
        metadata.validate_issuer()

    #: fragment
    metadata = AuthorizationServerMetadata({"issuer": "https://provider.test/#a=b"})
    with pytest.raises(ValueError, match="fragment"):
        metadata.validate_issuer()

    metadata = AuthorizationServerMetadata({"issuer": "https://provider.test/"})
    metadata.validate_issuer()


def test_validate_authorization_endpoint():
    # https
    metadata = AuthorizationServerMetadata(
        {"authorization_endpoint": "http://provider.test/"}
    )
    with pytest.raises(ValueError, match="https"):
        metadata.validate_authorization_endpoint()

    # valid https
    metadata = AuthorizationServerMetadata(
        {"authorization_endpoint": "https://provider.test/"}
    )
    metadata.validate_authorization_endpoint()

    # missing
    metadata = AuthorizationServerMetadata()
    with pytest.raises(ValueError, match="required"):
        metadata.validate_authorization_endpoint()

    # valid missing
    metadata = AuthorizationServerMetadata({"grant_types_supported": ["password"]})
    metadata.validate_authorization_endpoint()


def test_validate_token_endpoint():
    # implicit
    metadata = AuthorizationServerMetadata({"grant_types_supported": ["implicit"]})
    metadata.validate_token_endpoint()

    # missing
    metadata = AuthorizationServerMetadata()
    with pytest.raises(ValueError, match="required"):
        metadata.validate_token_endpoint()

    # https
    metadata = AuthorizationServerMetadata({"token_endpoint": "http://provider.test/"})
    with pytest.raises(ValueError, match="https"):
        metadata.validate_token_endpoint()

    # valid
    metadata = AuthorizationServerMetadata({"token_endpoint": "https://provider.test/"})
    metadata.validate_token_endpoint()


def test_validate_jwks_uri():
    # can missing
    metadata = AuthorizationServerMetadata()
    metadata.validate_jwks_uri()

    metadata = AuthorizationServerMetadata(
        {"jwks_uri": "http://provider.test/jwks.json"}
    )
    with pytest.raises(ValueError, match="https"):
        metadata.validate_jwks_uri()

    metadata = AuthorizationServerMetadata(
        {"jwks_uri": "https://provider.test/jwks.json"}
    )
    metadata.validate_jwks_uri()


def test_validate_registration_endpoint():
    metadata = AuthorizationServerMetadata()
    metadata.validate_registration_endpoint()

    metadata = AuthorizationServerMetadata(
        {"registration_endpoint": "http://provider.test/"}
    )
    with pytest.raises(ValueError, match="https"):
        metadata.validate_registration_endpoint()

    metadata = AuthorizationServerMetadata(
        {"registration_endpoint": "https://provider.test/"}
    )
    metadata.validate_registration_endpoint()


def test_validate_scopes_supported():
    metadata = AuthorizationServerMetadata()
    metadata.validate_scopes_supported()

    # not array
    metadata = AuthorizationServerMetadata({"scopes_supported": "foo"})
    with pytest.raises(ValueError, match="JSON array"):
        metadata.validate_scopes_supported()

    # valid
    metadata = AuthorizationServerMetadata({"scopes_supported": ["foo"]})
    metadata.validate_scopes_supported()


def test_validate_response_types_supported():
    # missing
    metadata = AuthorizationServerMetadata()
    with pytest.raises(ValueError, match="required"):
        metadata.validate_response_types_supported()

    # not array
    metadata = AuthorizationServerMetadata({"response_types_supported": "code"})
    with pytest.raises(ValueError, match="JSON array"):
        metadata.validate_response_types_supported()

    # valid
    metadata = AuthorizationServerMetadata({"response_types_supported": ["code"]})
    metadata.validate_response_types_supported()


def test_validate_response_modes_supported():
    metadata = AuthorizationServerMetadata()
    metadata.validate_response_modes_supported()

    # not array
    metadata = AuthorizationServerMetadata({"response_modes_supported": "query"})
    with pytest.raises(ValueError, match="JSON array"):
        metadata.validate_response_modes_supported()

    # valid
    metadata = AuthorizationServerMetadata({"response_modes_supported": ["query"]})
    metadata.validate_response_modes_supported()


def test_validate_grant_types_supported():
    metadata = AuthorizationServerMetadata()
    metadata.validate_grant_types_supported()

    # not array
    metadata = AuthorizationServerMetadata({"grant_types_supported": "password"})
    with pytest.raises(ValueError, match="JSON array"):
        metadata.validate_grant_types_supported()

    # valid
    metadata = AuthorizationServerMetadata({"grant_types_supported": ["password"]})
    metadata.validate_grant_types_supported()


def test_validate_token_endpoint_auth_methods_supported():
    metadata = AuthorizationServerMetadata()
    metadata.validate_token_endpoint_auth_methods_supported()

    # not array
    metadata = AuthorizationServerMetadata(
        {"token_endpoint_auth_methods_supported": "client_secret_basic"}
    )
    with pytest.raises(ValueError, match="JSON array"):
        metadata.validate_token_endpoint_auth_methods_supported()

    # valid
    metadata = AuthorizationServerMetadata(
        {"token_endpoint_auth_methods_supported": ["client_secret_basic"]}
    )
    metadata.validate_token_endpoint_auth_methods_supported()


def test_validate_token_endpoint_auth_signing_alg_values_supported():
    metadata = AuthorizationServerMetadata()
    metadata.validate_token_endpoint_auth_signing_alg_values_supported()

    metadata = AuthorizationServerMetadata(
        {"token_endpoint_auth_methods_supported": ["client_secret_jwt"]}
    )
    with pytest.raises(ValueError, match="required"):
        metadata.validate_token_endpoint_auth_signing_alg_values_supported()

    metadata = AuthorizationServerMetadata(
        {"token_endpoint_auth_signing_alg_values_supported": "RS256"}
    )
    with pytest.raises(ValueError, match="JSON array"):
        metadata.validate_token_endpoint_auth_signing_alg_values_supported()

    metadata = AuthorizationServerMetadata(
        {
            "token_endpoint_auth_methods_supported": ["client_secret_jwt"],
            "token_endpoint_auth_signing_alg_values_supported": ["RS256", "none"],
        }
    )
    with pytest.raises(ValueError, match="none"):
        metadata.validate_token_endpoint_auth_signing_alg_values_supported()


def test_validate_service_documentation():
    metadata = AuthorizationServerMetadata()
    metadata.validate_service_documentation()

    metadata = AuthorizationServerMetadata({"service_documentation": "invalid"})
    with pytest.raises(ValueError, match="MUST be a URL"):
        metadata.validate_service_documentation()

    metadata = AuthorizationServerMetadata(
        {"service_documentation": "https://provider.test/"}
    )
    metadata.validate_service_documentation()


def test_validate_ui_locales_supported():
    metadata = AuthorizationServerMetadata()
    metadata.validate_ui_locales_supported()

    # not array
    metadata = AuthorizationServerMetadata({"ui_locales_supported": "en"})
    with pytest.raises(ValueError, match="JSON array"):
        metadata.validate_ui_locales_supported()

    # valid
    metadata = AuthorizationServerMetadata({"ui_locales_supported": ["en"]})
    metadata.validate_ui_locales_supported()


def test_validate_op_policy_uri():
    metadata = AuthorizationServerMetadata()
    metadata.validate_op_policy_uri()

    metadata = AuthorizationServerMetadata({"op_policy_uri": "invalid"})
    with pytest.raises(ValueError, match="MUST be a URL"):
        metadata.validate_op_policy_uri()

    metadata = AuthorizationServerMetadata({"op_policy_uri": "https://provider.test/"})
    metadata.validate_op_policy_uri()


def test_validate_op_tos_uri():
    metadata = AuthorizationServerMetadata()
    metadata.validate_op_tos_uri()

    metadata = AuthorizationServerMetadata({"op_tos_uri": "invalid"})
    with pytest.raises(ValueError, match="MUST be a URL"):
        metadata.validate_op_tos_uri()

    metadata = AuthorizationServerMetadata({"op_tos_uri": "https://provider.test/"})
    metadata.validate_op_tos_uri()


def test_validate_revocation_endpoint():
    metadata = AuthorizationServerMetadata()
    metadata.validate_revocation_endpoint()

    # https
    metadata = AuthorizationServerMetadata(
        {"revocation_endpoint": "http://provider.test/"}
    )
    with pytest.raises(ValueError, match="https"):
        metadata.validate_revocation_endpoint()

    # valid
    metadata = AuthorizationServerMetadata(
        {"revocation_endpoint": "https://provider.test/"}
    )
    metadata.validate_revocation_endpoint()


def test_validate_revocation_endpoint_auth_methods_supported():
    metadata = AuthorizationServerMetadata()
    metadata.validate_revocation_endpoint_auth_methods_supported()

    # not array
    metadata = AuthorizationServerMetadata(
        {"revocation_endpoint_auth_methods_supported": "client_secret_basic"}
    )
    with pytest.raises(ValueError, match="JSON array"):
        metadata.validate_revocation_endpoint_auth_methods_supported()

    # valid
    metadata = AuthorizationServerMetadata(
        {"revocation_endpoint_auth_methods_supported": ["client_secret_basic"]}
    )
    metadata.validate_revocation_endpoint_auth_methods_supported()


def test_validate_revocation_endpoint_auth_signing_alg_values_supported():
    metadata = AuthorizationServerMetadata()
    metadata.validate_revocation_endpoint_auth_signing_alg_values_supported()

    metadata = AuthorizationServerMetadata(
        {"revocation_endpoint_auth_methods_supported": ["client_secret_jwt"]}
    )
    with pytest.raises(ValueError, match="required"):
        metadata.validate_revocation_endpoint_auth_signing_alg_values_supported()

    metadata = AuthorizationServerMetadata(
        {"revocation_endpoint_auth_signing_alg_values_supported": "RS256"}
    )
    with pytest.raises(ValueError, match="JSON array"):
        metadata.validate_revocation_endpoint_auth_signing_alg_values_supported()

    metadata = AuthorizationServerMetadata(
        {
            "revocation_endpoint_auth_methods_supported": ["client_secret_jwt"],
            "revocation_endpoint_auth_signing_alg_values_supported": [
                "RS256",
                "none",
            ],
        }
    )
    with pytest.raises(ValueError, match="none"):
        metadata.validate_revocation_endpoint_auth_signing_alg_values_supported()


def test_validate_introspection_endpoint():
    metadata = AuthorizationServerMetadata()
    metadata.validate_introspection_endpoint()

    # https
    metadata = AuthorizationServerMetadata(
        {"introspection_endpoint": "http://provider.test/"}
    )
    with pytest.raises(ValueError, match="https"):
        metadata.validate_introspection_endpoint()

    # valid
    metadata = AuthorizationServerMetadata(
        {"introspection_endpoint": "https://provider.test/"}
    )
    metadata.validate_introspection_endpoint()


def test_validate_introspection_endpoint_auth_methods_supported():
    metadata = AuthorizationServerMetadata()
    metadata.validate_introspection_endpoint_auth_methods_supported()

    # not array
    metadata = AuthorizationServerMetadata(
        {"introspection_endpoint_auth_methods_supported": "client_secret_basic"}
    )
    with pytest.raises(ValueError, match="JSON array"):
        metadata.validate_introspection_endpoint_auth_methods_supported()

    # valid
    metadata = AuthorizationServerMetadata(
        {"introspection_endpoint_auth_methods_supported": ["client_secret_basic"]}
    )
    metadata.validate_introspection_endpoint_auth_methods_supported()


def test_validate_introspection_endpoint_auth_signing_alg_values_supported():
    metadata = AuthorizationServerMetadata()
    metadata.validate_introspection_endpoint_auth_signing_alg_values_supported()

    metadata = AuthorizationServerMetadata(
        {"introspection_endpoint_auth_methods_supported": ["client_secret_jwt"]}
    )
    with pytest.raises(ValueError, match="required"):
        metadata.validate_introspection_endpoint_auth_signing_alg_values_supported()

    metadata = AuthorizationServerMetadata(
        {"introspection_endpoint_auth_signing_alg_values_supported": "RS256"}
    )
    with pytest.raises(ValueError, match="JSON array"):
        metadata.validate_introspection_endpoint_auth_signing_alg_values_supported()

    metadata = AuthorizationServerMetadata(
        {
            "introspection_endpoint_auth_methods_supported": ["client_secret_jwt"],
            "introspection_endpoint_auth_signing_alg_values_supported": [
                "RS256",
                "none",
            ],
        }
    )
    with pytest.raises(ValueError, match="none"):
        metadata.validate_introspection_endpoint_auth_signing_alg_values_supported()


def test_validate_code_challenge_methods_supported():
    metadata = AuthorizationServerMetadata()
    metadata.validate_code_challenge_methods_supported()

    # not array
    metadata = AuthorizationServerMetadata({"code_challenge_methods_supported": "S256"})
    with pytest.raises(ValueError, match="JSON array"):
        metadata.validate_code_challenge_methods_supported()

    # valid
    metadata = AuthorizationServerMetadata(
        {"code_challenge_methods_supported": ["S256"]}
    )
    metadata.validate_code_challenge_methods_supported()



import logging

from authlib.jose import JoseError
from authlib.jose import jwt

from ..rfc6749 import BaseGrant
from ..rfc6749 import InvalidClientError
from ..rfc6749 import InvalidGrantError
from ..rfc6749 import InvalidRequestError
from ..rfc6749 import TokenEndpointMixin
from ..rfc6749 import UnauthorizedClientError
from .assertion import sign_jwt_bearer_assertion

log = logging.getLogger(__name__)
JWT_BEARER_GRANT_TYPE = "urn:ietf:params:oauth:grant-type:jwt-bearer"


class JWTBearerGrant(BaseGrant, TokenEndpointMixin):
    GRANT_TYPE = JWT_BEARER_GRANT_TYPE

    #: Options for verifying JWT payload claims. Developers MAY
    #: overwrite this constant to create a more strict options.
    CLAIMS_OPTIONS = {
        "iss": {"essential": True},
        "aud": {"essential": True},
        "exp": {"essential": True},
    }

    # A small allowance of time, typically no more than a few minutes,
    # to account for clock skew. The default is 60 seconds.
    LEEWAY = 60

    @staticmethod
    def sign(
        key,
        issuer,
        audience,
        subject=None,
        issued_at=None,
        expires_at=None,
        claims=None,
        **kwargs,
    ):
        return sign_jwt_bearer_assertion(
            key, issuer, audience, subject, issued_at, expires_at, claims, **kwargs
        )

    def process_assertion_claims(self, assertion):
        """Extract JWT payload claims from request "assertion", per
        `Section 3.1`_.

        :param assertion: assertion string value in the request
        :return: JWTClaims
        :raise: InvalidGrantError

        .. _`Section 3.1`: https://tools.ietf.org/html/rfc7523#section-3.1
        """
        try:
            claims = jwt.decode(
                assertion, self.resolve_public_key, claims_options=self.CLAIMS_OPTIONS
            )
            claims.validate(leeway=self.LEEWAY)
        except JoseError as e:
            log.debug("Assertion Error: %r", e)
            raise InvalidGrantError(description=e.description) from e
        return claims

    def resolve_public_key(self, headers, payload):
        client = self.resolve_issuer_client(payload["iss"])
        return self.resolve_client_key(client, headers, payload)

    def validate_token_request(self):
        """The client makes a request to the token endpoint by sending the
        following parameters using the "application/x-www-form-urlencoded"
        format per `Section 2.1`_:

        grant_type
             REQUIRED.  Value MUST be set to
             "urn:ietf:params:oauth:grant-type:jwt-bearer".

        assertion
             REQUIRED.  Value MUST contain a single JWT.

        scope
            OPTIONAL.

        The following example demonstrates an access token request with a JWT
        as an authorization grant:

        .. code-block:: http

            POST /token.oauth2 HTTP/1.1
            Host: as.example.com
            Content-Type: application/x-www-form-urlencoded

            grant_type=urn%3Aietf%3Aparams%3Aoauth%3Agrant-type%3Ajwt-bearer
            &assertion=eyJhbGciOiJFUzI1NiIsImtpZCI6IjE2In0.
            eyJpc3Mi[...omitted for brevity...].
            J9l-ZhwP[...omitted for brevity...]

        .. _`Section 2.1`: https://tools.ietf.org/html/rfc7523#section-2.1
        """
        assertion = self.request.form.get("assertion")
        if not assertion:
            raise InvalidRequestError("Missing 'assertion' in request")

        claims = self.process_assertion_claims(assertion)
        client = self.resolve_issuer_client(claims["iss"])
        log.debug("Validate token request of %s", client)

        if not client.check_grant_type(self.GRANT_TYPE):
            raise UnauthorizedClientError(
                f"The client is not authorized to use 'grant_type={self.GRANT_TYPE}'"
            )

        self.request.client = client
        self.validate_requested_scope()

        subject = claims.get("sub")
        if subject:
            user = self.authenticate_user(subject)
            if not user:
                raise InvalidGrantError(description="Invalid 'sub' value in assertion")

            log.debug("Check client(%s) permission to User(%s)", client, user)
            if not self.has_granted_permission(client, user):
                raise InvalidClientError(
                    description="Client has no permission to access user data"
                )
            self.request.user = user

    def create_token_response(self):
        """If valid and authorized, the authorization server issues an access
        token.
        """
        token = self.generate_token(
            scope=self.request.payload.scope,
            user=self.request.user,
            include_refresh_token=False,
        )
        log.debug("Issue token %r to %r", token, self.request.client)
        self.save_token(token)
        return 200, token, self.TOKEN_RESPONSE_HEADER

    def resolve_issuer_client(self, issuer):
        """Fetch client via "iss" in assertion claims. Developers MUST
        implement this method in subclass, e.g.::

            def resolve_issuer_client(self, issuer):
                return Client.query_by_iss(issuer)

        :param issuer: "iss" value in assertion
        :return: Client instance
        """
        raise NotImplementedError()

    def resolve_client_key(self, client, headers, payload):
        """Resolve client key to decode assertion data. Developers MUST
        implement this method in subclass. For instance, there is a
        "jwks" column on client table, e.g.::

            def resolve_client_key(self, client, headers, payload):
                # from authlib.jose import JsonWebKey

                key_set = JsonWebKey.import_key_set(client.jwks)
                return key_set.find_by_kid(headers["kid"])

        :param client: instance of OAuth client model
        :param headers: headers part of the JWT
        :param payload: payload part of the JWT
        :return: ``authlib.jose.Key`` instance
        """
        raise NotImplementedError()

    def authenticate_user(self, subject):
        """Authenticate user with the given assertion claims. Developers MUST
        implement it in subclass, e.g.::

            def authenticate_user(self, subject):
                return User.get_by_sub(subject)

        :param subject: "sub" value in claims
        :return: User instance
        """
        raise NotImplementedError()

    def has_granted_permission(self, client, user):
        """Check if the client has permission to access the given user's resource.
        Developers MUST implement it in subclass, e.g.::

            def has_granted_permission(self, client, user):
                permission = ClientUserGrant.query(client=client, user=user)
                return permission.granted

        :param client: instance of OAuth client model
        :param user: instance of User model
        :return: bool
        """
        raise NotImplementedError()



REQUESTS_SESSION_KWARGS = [
    "proxies",
    "hooks",
    "stream",
    "verify",
    "cert",
    "max_redirects",
    "trust_env",
]


def update_session_configure(session, kwargs):
    for k in REQUESTS_SESSION_KWARGS:
        if k in kwargs:
            setattr(session, k, kwargs.pop(k))



from authlib.common.errors import AuthlibBaseError


class OAuthError(AuthlibBaseError):
    error = "oauth_error"


class MissingRequestTokenError(OAuthError):
    error = "missing_request_token"


class MissingTokenError(OAuthError):
    error = "missing_token"


class TokenExpiredError(OAuthError):
    error = "token_expired"


class InvalidTokenError(OAuthError):
    error = "token_invalid"


class UnsupportedTokenTypeError(OAuthError):
    error = "unsupported_token_type"


class MismatchingStateError(OAuthError):
    error = "mismatching_state"
    description = "CSRF Warning! State not equal in request and response."



from authlib.jose import JsonWebSignature
from authlib.jose import JsonWebToken
from authlib.jose.errors import JoseError

from ..rfc6749 import AuthorizationServer
from ..rfc6749 import ClientMixin
from ..rfc6749 import InvalidRequestError
from ..rfc6749.authenticate_client import _validate_client
from ..rfc6749.requests import BasicOAuth2Payload
from ..rfc6749.requests import OAuth2Request
from .errors import InvalidRequestObjectError
from .errors import InvalidRequestUriError
from .errors import RequestNotSupportedError
from .errors import RequestUriNotSupportedError


class JWTAuthenticationRequest:
    """Authorization server extension implementing the support
    for JWT secured authentication request, as defined in :rfc:`RFC9101 <9101>`.

    :param support_request: Whether to enable support for the ``request`` parameter.
    :param support_request_uri: Whether to enable support for the ``request_uri`` parameter.

    This extension is intended to be inherited and registered into the authorization server::

        class JWTAuthenticationRequest(rfc9101.JWTAuthenticationRequest):
            def resolve_client_public_key(self, client: ClientMixin):
                return get_jwks_for_client(client)

            def get_request_object(self, request_uri: str):
                try:
                    return requests.get(request_uri).text
                except requests.Exception:
                    return None

            def get_server_metadata(self):
                return {
                    "issuer": ...,
                    "authorization_endpoint": ...,
                    "require_signed_request_object": ...,
                }

            def get_client_require_signed_request_object(self, client: ClientMixin):
                return client.require_signed_request_object


        authorization_server.register_extension(JWTAuthenticationRequest())
    """

    def __init__(self, support_request: bool = True, support_request_uri: bool = True):
        self.support_request = support_request
        self.support_request_uri = support_request_uri

    def __call__(self, authorization_server: AuthorizationServer):
        authorization_server.register_hook(
            "before_get_authorization_grant", self.parse_authorization_request
        )

    def parse_authorization_request(
        self, authorization_server: AuthorizationServer, request: OAuth2Request
    ):
        client = _validate_client(
            authorization_server.query_client, request.payload.client_id
        )
        if not self._shoud_proceed_with_request_object(
            authorization_server, request, client
        ):
            return

        raw_request_object = self._get_raw_request_object(authorization_server, request)
        request_object = self._decode_request_object(
            request, client, raw_request_object
        )
        payload = BasicOAuth2Payload(request_object)
        request.payload = payload

    def _shoud_proceed_with_request_object(
        self,
        authorization_server: AuthorizationServer,
        request: OAuth2Request,
        client: ClientMixin,
    ) -> bool:
        if "request" in request.payload.data and "request_uri" in request.payload.data:
            raise InvalidRequestError(
                "The 'request' and 'request_uri' parameters are mutually exclusive.",
                state=request.payload.state,
            )

        if "request" in request.payload.data:
            if not self.support_request:
                raise RequestNotSupportedError(state=request.payload.state)
            return True

        if "request_uri" in request.payload.data:
            if not self.support_request_uri:
                raise RequestUriNotSupportedError(state=request.payload.state)
            return True

        # When the value of it [require_signed_request_object] as client metadata is true,
        # then the server MUST reject the authorization request
        # from the client that does not conform to this specification.
        if self.get_client_require_signed_request_object(client):
            raise InvalidRequestError(
                "Authorization requests for this client must use signed request objects.",
                state=request.payload.state,
            )

        # When the value of it [require_signed_request_object] as server metadata is true,
        # then the server MUST reject the authorization request
        # from any client that does not conform to this specification.
        metadata = self.get_server_metadata()
        if metadata and metadata.get("require_signed_request_object", False):
            raise InvalidRequestError(
                "Authorization requests for this server must use signed request objects.",
                state=request.payload.state,
            )

        return False

    def _get_raw_request_object(
        self, authorization_server: AuthorizationServer, request: OAuth2Request
    ) -> str:
        if "request_uri" in request.payload.data:
            raw_request_object = self.get_request_object(
                request.payload.data["request_uri"]
            )
            if not raw_request_object:
                raise InvalidRequestUriError(state=request.payload.state)

        else:
            raw_request_object = request.payload.data["request"]

        return raw_request_object

    def _decode_request_object(
        self, request, client: ClientMixin, raw_request_object: str
    ):
        jwks = self.resolve_client_public_key(client)
        try:
            jwt = JsonWebToken(list(JsonWebSignature.ALGORITHMS_REGISTRY.keys()))
            request_object = jwt.decode(raw_request_object, jwks)
            request_object.validate()

        except JoseError as error:
            raise InvalidRequestObjectError(
                description=error.description or InvalidRequestObjectError.description,
                state=request.payload.state,
            ) from error

        # It MUST also reject the request if the Request Object uses an
        # alg value of none when this server metadata value is true.
        # If omitted, the default value is false.
        if (
            self.get_client_require_signed_request_object(client)
            and request_object.header["alg"] == "none"
        ):
            raise InvalidRequestError(
                "Authorization requests for this client must use signed request objects.",
                state=request.payload.state,
            )

        # It MUST also reject the request if the Request Object uses an
        # alg value of none. If omitted, the default value is false.
        metadata = self.get_server_metadata()
        if (
            metadata
            and metadata.get("require_signed_request_object", False)
            and request_object.header["alg"] == "none"
        ):
            raise InvalidRequestError(
                "Authorization requests for this server must use signed request objects.",
                state=request.payload.state,
            )

        # The client ID values in the client_id request parameter and in
        # the Request Object client_id claim MUST be identical.
        if request_object["client_id"] != request.payload.client_id:
            raise InvalidRequestError(
                "The 'client_id' claim from the request parameters "
                "and the request object claims don't match.",
                state=request.payload.state,
            )

        # The Request Object MAY be sent by value, as described in Section 5.1,
        # or by reference, as described in Section 5.2. request and
        # request_uri parameters MUST NOT be included in Request Objects.
        if "request" in request_object or "request_uri" in request_object:
            raise InvalidRequestError(
                "The 'request' and 'request_uri' parameters must not be included in the request object.",
                state=request.payload.state,
            )

        return request_object

    def get_request_object(self, request_uri: str):
        """Download the request object at ``request_uri``.

        This method must be implemented if the ``request_uri`` parameter is supported::

            class JWTAuthenticationRequest(rfc9101.JWTAuthenticationRequest):
                def get_request_object(self, request_uri: str):
                    try:
                        return requests.get(request_uri).text
                    except requests.Exception:
                        return None
        """
        raise NotImplementedError()

    def resolve_client_public_keys(self, client: ClientMixin):
        """Resolve the client public key for verifying the JWT signature.
        A client may have many public keys, in this case, we can retrieve it
        via ``kid`` value in headers. Developers MUST implement this method::

            class JWTAuthenticationRequest(rfc9101.JWTAuthenticationRequest):
                def resolve_client_public_key(self, client):
                    if client.jwks_uri:
                        return requests.get(client.jwks_uri).json

                    return client.jwks
        """
        raise NotImplementedError()

    def get_server_metadata(self) -> dict:
        """Return server metadata which includes supported grant types,
        response types and etc.

        When the ``require_signed_request_object`` claim is :data:`True`,
        all clients require that authorization requests
        use request objects, and an error will be returned when the authorization
        request payload is passed in the request body or query string::

            class JWTAuthenticationRequest(rfc9101.JWTAuthenticationRequest):
                def get_server_metadata(self):
                    return {
                        "issuer": ...,
                        "authorization_endpoint": ...,
                        "require_signed_request_object": ...,
                    }

        """
        return {}  # pragma: no cover

    def get_client_require_signed_request_object(self, client: ClientMixin) -> bool:
        """Return the 'require_signed_request_object' client metadata.

        When :data:`True`, the client requires that authorization requests
        use request objects, and an error will be returned when the authorization
        request payload is passed in the request body or query string::

           class JWTAuthenticationRequest(rfc9101.JWTAuthenticationRequest):
               def get_client_require_signed_request_object(self, client):
                   return client.require_signed_request_object

        If not implemented, the value is considered as :data:`False`.
        """
        return False  # pragma: no cover



import json

import pytest

from authlib.oauth2.rfc6749.grants import (
    ResourceOwnerPasswordCredentialsGrant as _PasswordGrant,
)

from .models import Client
from .models import User
from .oauth2_server import create_basic_auth


@pytest.fixture(autouse=True)
def server(server):
    class PasswordGrant(_PasswordGrant):
        def authenticate_user(self, username, password):
            try:
                user = User.objects.get(username=username)
                if user.check_password(password):
                    return user
            except User.DoesNotExist:
                return None

    server.register_grant(PasswordGrant)
    return server


@pytest.fixture(autouse=True)
def client(user):
    client = Client(
        user_id=user.pk,
        client_id="client-id",
        client_secret="client-secret",
        scope="",
        grant_type="password",
        token_endpoint_auth_method="client_secret_basic",
        default_redirect_uri="https://client.test",
    )
    client.save()
    yield client
    client.delete()


def test_invalid_client(factory, server):
    request = factory.post(
        "/oauth/token",
        data={"grant_type": "password", "username": "foo", "password": "ok"},
    )
    resp = server.create_token_response(request)
    assert resp.status_code == 401
    data = json.loads(resp.content)
    assert data["error"] == "invalid_client"

    request = factory.post(
        "/oauth/token",
        data={"grant_type": "password", "username": "foo", "password": "ok"},
        HTTP_AUTHORIZATION=create_basic_auth("invalid", "client-secret"),
    )
    resp = server.create_token_response(request)
    assert resp.status_code == 401
    data = json.loads(resp.content)
    assert data["error"] == "invalid_client"


def test_invalid_scope(factory, server):
    server.scopes_supported = ["profile"]
    request = factory.post(
        "/oauth/token",
        data={
            "grant_type": "password",
            "username": "foo",
            "password": "ok",
            "scope": "invalid",
        },
        HTTP_AUTHORIZATION=create_basic_auth("client-id", "client-secret"),
    )
    resp = server.create_token_response(request)
    assert resp.status_code == 400
    data = json.loads(resp.content)
    assert data["error"] == "invalid_scope"


def test_invalid_request(factory, server):
    auth_header = create_basic_auth("client-id", "client-secret")

    # case 1
    request = factory.get(
        "/oauth/token?grant_type=password",
        HTTP_AUTHORIZATION=auth_header,
    )
    resp = server.create_token_response(request)
    assert resp.status_code == 400
    data = json.loads(resp.content)
    assert data["error"] == "unsupported_grant_type"

    # case 2
    request = factory.post(
        "/oauth/token",
        data={"grant_type": "password"},
        HTTP_AUTHORIZATION=auth_header,
    )
    resp = server.create_token_response(request)
    assert resp.status_code == 400
    data = json.loads(resp.content)
    assert data["error"] == "invalid_request"

    # case 3
    request = factory.post(
        "/oauth/token",
        data={"grant_type": "password", "username": "foo"},
        HTTP_AUTHORIZATION=auth_header,
    )
    resp = server.create_token_response(request)
    assert resp.status_code == 400
    data = json.loads(resp.content)
    assert data["error"] == "invalid_request"

    # case 4
    request = factory.post(
        "/oauth/token",
        data={
            "grant_type": "password",
            "username": "foo",
            "password": "wrong",
        },
        HTTP_AUTHORIZATION=auth_header,
    )
    resp = server.create_token_response(request)
    assert resp.status_code == 400
    data = json.loads(resp.content)
    assert data["error"] == "invalid_request"


def test_unauthorized_client(factory, server, client):
    client.grant_type = "invalid"
    client.save()
    request = factory.post(
        "/oauth/token",
        data={
            "grant_type": "password",
            "username": "foo",
            "password": "ok",
        },
        HTTP_AUTHORIZATION=create_basic_auth("client-id", "client-secret"),
    )
    resp = server.create_token_response(request)
    assert resp.status_code == 400
    data = json.loads(resp.content)
    assert data["error"] == "unauthorized_client"


def test_authorize_token(factory, server):
    request = factory.post(
        "/oauth/token",
        data={
            "grant_type": "password",
            "username": "foo",
            "password": "ok",
        },
        HTTP_AUTHORIZATION=create_basic_auth("client-id", "client-secret"),
    )
    resp = server.create_token_response(request)
    assert resp.status_code == 200
    data = json.loads(resp.content)
    assert "access_token" in data



import logging

from authlib.common.security import generate_token
from authlib.oauth2.rfc6749 import InvalidScopeError
from authlib.oauth2.rfc6749.grants.authorization_code import (
    validate_code_authorization_request,
)

from .implicit import OpenIDImplicitGrant
from .util import is_openid_scope
from .util import validate_nonce

log = logging.getLogger(__name__)


class OpenIDHybridGrant(OpenIDImplicitGrant):
    #: Generated "code" length
    AUTHORIZATION_CODE_LENGTH = 48

    RESPONSE_TYPES = {"code id_token", "code token", "code id_token token"}
    GRANT_TYPE = "code"
    DEFAULT_RESPONSE_MODE = "fragment"

    def generate_authorization_code(self):
        """ "The method to generate "code" value for authorization code data.
        Developers may rewrite this method, or customize the code length with::

            class MyAuthorizationCodeGrant(AuthorizationCodeGrant):
                AUTHORIZATION_CODE_LENGTH = 32  # default is 48
        """
        return generate_token(self.AUTHORIZATION_CODE_LENGTH)

    def save_authorization_code(self, code, request):
        """Save authorization_code for later use. Developers MUST implement
        it in subclass. Here is an example::

            def save_authorization_code(self, code, request):
                client = request.client
                auth_code = AuthorizationCode(
                    code=code,
                    client_id=client.client_id,
                    redirect_uri=request.payload.redirect_uri,
                    scope=request.payload.scope,
                    nonce=request.payload.data.get("nonce"),
                    user_id=request.user.id,
                )
                auth_code.save()
        """
        raise NotImplementedError()

    def validate_authorization_request(self):
        if not is_openid_scope(self.request.payload.scope):
            raise InvalidScopeError(
                "Missing 'openid' scope",
                redirect_uri=self.request.payload.redirect_uri,
                redirect_fragment=True,
            )
        self.register_hook(
            "after_validate_authorization_request_payload",
            lambda grant, redirect_uri: validate_nonce(
                grant.request, grant.exists_nonce, required=True
            ),
        )
        return validate_code_authorization_request(self)

    def create_granted_params(self, grant_user):
        self.request.user = grant_user
        client = self.request.client
        code = self.generate_authorization_code()
        self.save_authorization_code(code, self.request)
        params = [("code", code)]
        token = self.generate_token(
            grant_type="implicit",
            user=grant_user,
            scope=self.request.payload.scope,
            include_refresh_token=False,
        )

        response_types = self.request.payload.response_type.split()
        if "token" in response_types:
            log.debug("Grant token %r to %r", token, client)
            self.server.save_token(token, self.request)
            if "id_token" in response_types:
                token = self.process_implicit_token(token, code)
        else:
            # response_type is "code id_token"
            token = {"expires_in": token["expires_in"], "scope": token["scope"]}
            token = self.process_implicit_token(token, code)

        params.extend([(k, token[k]) for k in token])
        return params



from authlib.common.urls import urlparse


def get_well_known_url(issuer, external=False, suffix="oauth-authorization-server"):
    """Get well-known URI with issuer via `Section 3.1`_.

    .. _`Section 3.1`: https://tools.ietf.org/html/rfc8414#section-3.1

    :param issuer: URL of the issuer
    :param external: return full external url or not
    :param suffix: well-known URI suffix for RFC8414
    :return: URL
    """
    parsed = urlparse.urlparse(issuer)
    path = parsed.path
    if path and path != "/":
        url_path = f"/.well-known/{suffix}{path}"
    else:
        url_path = f"/.well-known/{suffix}"
    if not external:
        return url_path
    return parsed.scheme + "://" + parsed.netloc + url_path



import pytest
from flask import json

from authlib.oauth2.rfc7523 import JWTBearerGrant as _JWTBearerGrant
from authlib.oauth2.rfc7523 import JWTBearerTokenGenerator
from tests.util import read_file_path

from .models import Client
from .models import db


class JWTBearerGrant(_JWTBearerGrant):
    def resolve_issuer_client(self, issuer):
        return Client.query.filter_by(client_id=issuer).first()

    def resolve_client_key(self, client, headers, payload):
        keys = {"1": "foo", "2": "bar"}
        return keys[headers["kid"]]

    def authenticate_user(self, subject):
        return None

    def has_granted_permission(self, client, user):
        return True


@pytest.fixture(autouse=True)
def server(server):
    server.register_grant(JWTBearerGrant)
    return server


@pytest.fixture(autouse=True)
def client(client, db):
    client.set_client_metadata(
        {
            "scope": "profile",
            "redirect_uris": ["https://client.test/authorized"],
            "grant_types": [JWTBearerGrant.GRANT_TYPE],
        }
    )
    db.session.add(client)
    db.session.commit()
    return client


def test_missing_assertion(test_client):
    rv = test_client.post(
        "/oauth/token", data={"grant_type": JWTBearerGrant.GRANT_TYPE}
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_request"
    assert "assertion" in resp["error_description"]


def test_invalid_assertion(test_client):
    assertion = JWTBearerGrant.sign(
        "foo",
        issuer="client-id",
        audience="https://provider.test/token",
        subject="none",
        header={"alg": "HS256", "kid": "1"},
    )
    rv = test_client.post(
        "/oauth/token",
        data={"grant_type": JWTBearerGrant.GRANT_TYPE, "assertion": assertion},
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_grant"


def test_authorize_token(test_client):
    assertion = JWTBearerGrant.sign(
        "foo",
        issuer="client-id",
        audience="https://provider.test/token",
        subject=None,
        header={"alg": "HS256", "kid": "1"},
    )
    rv = test_client.post(
        "/oauth/token",
        data={"grant_type": JWTBearerGrant.GRANT_TYPE, "assertion": assertion},
    )
    resp = json.loads(rv.data)
    assert "access_token" in resp


def test_unauthorized_client(test_client, client):
    client.set_client_metadata(
        {
            "scope": "profile",
            "redirect_uris": ["https://client.test/authorized"],
            "grant_types": ["password"],
        }
    )
    db.session.add(client)
    db.session.commit()

    assertion = JWTBearerGrant.sign(
        "bar",
        issuer="client-id",
        audience="https://provider.test/token",
        subject=None,
        header={"alg": "HS256", "kid": "2"},
    )
    rv = test_client.post(
        "/oauth/token",
        data={"grant_type": JWTBearerGrant.GRANT_TYPE, "assertion": assertion},
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "unauthorized_client"


def test_token_generator(test_client, app, server):
    m = "tests.flask.test_oauth2.oauth2_server:token_generator"
    app.config.update({"OAUTH2_ACCESS_TOKEN_GENERATOR": m})
    server.load_config(app.config)
    assertion = JWTBearerGrant.sign(
        "foo",
        issuer="client-id",
        audience="https://provider.test/token",
        subject=None,
        header={"alg": "HS256", "kid": "1"},
    )
    rv = test_client.post(
        "/oauth/token",
        data={"grant_type": JWTBearerGrant.GRANT_TYPE, "assertion": assertion},
    )
    resp = json.loads(rv.data)
    assert "access_token" in resp
    assert "c-" in resp["access_token"]


def test_jwt_bearer_token_generator(test_client, server):
    private_key = read_file_path("jwks_private.json")
    server.register_token_generator(
        JWTBearerGrant.GRANT_TYPE, JWTBearerTokenGenerator(private_key)
    )
    assertion = JWTBearerGrant.sign(
        "foo",
        issuer="client-id",
        audience="https://provider.test/token",
        subject=None,
        header={"alg": "HS256", "kid": "1"},
    )
    rv = test_client.post(
        "/oauth/token",
        data={"grant_type": JWTBearerGrant.GRANT_TYPE, "assertion": assertion},
    )
    resp = json.loads(rv.data)
    assert "access_token" in resp
    assert resp["access_token"].count(".") == 2



import json
import time
from collections.abc import Hashable
from typing import Any
from typing import Optional

from ..base_client import FrameworkIntegration


class StarletteIntegration(FrameworkIntegration):
    async def _get_cache_data(self, key: Hashable):
        value = await self.cache.get(key)
        if not value:
            return None
        try:
            return json.loads(value)
        except (TypeError, ValueError):
            return None

    async def get_state_data(
        self, session: Optional[dict[str, Any]], state: str
    ) -> dict[str, Any]:
        key = f"_state_{self.name}_{state}"
        if self.cache:
            value = await self._get_cache_data(key)
        elif session is not None:
            value = session.get(key)
        else:
            value = None

        if value:
            return value.get("data")
        return None

    async def set_state_data(
        self, session: Optional[dict[str, Any]], state: str, data: Any
    ):
        key_prefix = f"_state_{self.name}_"
        key = f"{key_prefix}{state}"
        if self.cache:
            await self.cache.set(key, json.dumps({"data": data}), self.expires_in)
        elif session is not None:
            # clear old state data to avoid session size growing
            for old_key in list(session.keys()):
                if old_key.startswith(key_prefix):
                    session.pop(old_key)
            now = time.time()
            session[key] = {"data": data, "exp": now + self.expires_in}

    async def clear_state_data(self, session: Optional[dict[str, Any]], state: str):
        key = f"_state_{self.name}_{state}"
        if self.cache:
            await self.cache.delete(key)
        elif session is not None:
            session.pop(key, None)
            self._clear_session_state(session)

    def update_token(self, token, refresh_token=None, access_token=None):
        pass

    @staticmethod
    def load_config(oauth, name, params):
        if not oauth.config:
            return {}

        rv = {}
        for k in params:
            conf_key = f"{name}_{k}".upper()
            v = oauth.config.get(conf_key, default=None)
            if v is not None:
                rv[k] = v
        return rv



import httpx
from httpx import USE_CLIENT_DEFAULT
from httpx import Response

from authlib.oauth2.rfc7521 import AssertionClient as _AssertionClient
from authlib.oauth2.rfc7523 import JWTBearerGrant

from ..base_client import OAuthError
from .oauth2_client import OAuth2Auth
from .utils import extract_client_kwargs

__all__ = ["AsyncAssertionClient"]


class AsyncAssertionClient(_AssertionClient, httpx.AsyncClient):
    token_auth_class = OAuth2Auth
    oauth_error_class = OAuthError
    JWT_BEARER_GRANT_TYPE = JWTBearerGrant.GRANT_TYPE
    ASSERTION_METHODS = {
        JWT_BEARER_GRANT_TYPE: JWTBearerGrant.sign,
    }
    DEFAULT_GRANT_TYPE = JWT_BEARER_GRANT_TYPE

    def __init__(
        self,
        token_endpoint,
        issuer,
        subject,
        audience=None,
        grant_type=None,
        claims=None,
        token_placement="header",
        scope=None,
        **kwargs,
    ):
        client_kwargs = extract_client_kwargs(kwargs)
        httpx.AsyncClient.__init__(self, **client_kwargs)

        _AssertionClient.__init__(
            self,
            session=None,
            token_endpoint=token_endpoint,
            issuer=issuer,
            subject=subject,
            audience=audience,
            grant_type=grant_type,
            claims=claims,
            token_placement=token_placement,
            scope=scope,
            **kwargs,
        )

    async def request(
        self, method, url, withhold_token=False, auth=USE_CLIENT_DEFAULT, **kwargs
    ) -> Response:
        """Send request with auto refresh token feature."""
        if not withhold_token and auth is USE_CLIENT_DEFAULT:
            if not self.token or self.token.is_expired():
                await self.refresh_token()

            auth = self.token_auth
        return await super().request(method, url, auth=auth, **kwargs)

    async def _refresh_token(self, data):
        resp = await self.request(
            "POST", self.token_endpoint, data=data, withhold_token=True
        )

        return self.parse_response_token(resp)


class AssertionClient(_AssertionClient, httpx.Client):
    token_auth_class = OAuth2Auth
    oauth_error_class = OAuthError
    JWT_BEARER_GRANT_TYPE = JWTBearerGrant.GRANT_TYPE
    ASSERTION_METHODS = {
        JWT_BEARER_GRANT_TYPE: JWTBearerGrant.sign,
    }
    DEFAULT_GRANT_TYPE = JWT_BEARER_GRANT_TYPE

    def __init__(
        self,
        token_endpoint,
        issuer,
        subject,
        audience=None,
        grant_type=None,
        claims=None,
        token_placement="header",
        scope=None,
        **kwargs,
    ):
        client_kwargs = extract_client_kwargs(kwargs)
        # app keyword was dropped!
        app_value = client_kwargs.pop("app", None)
        if app_value is not None:
            client_kwargs["transport"] = httpx.WSGITransport(app=app_value)

        httpx.Client.__init__(self, **client_kwargs)

        _AssertionClient.__init__(
            self,
            session=self,
            token_endpoint=token_endpoint,
            issuer=issuer,
            subject=subject,
            audience=audience,
            grant_type=grant_type,
            claims=claims,
            token_placement=token_placement,
            scope=scope,
            **kwargs,
        )

    def request(
        self, method, url, withhold_token=False, auth=USE_CLIENT_DEFAULT, **kwargs
    ):
        """Send request with auto refresh token feature."""
        if not withhold_token and auth is USE_CLIENT_DEFAULT:
            if not self.token or self.token.is_expired():
                self.refresh_token()

            auth = self.token_auth
        return super().request(method, url, auth=auth, **kwargs)



import time

import pytest
from httpx import WSGITransport

from authlib.integrations.httpx_client import AssertionClient

from ..wsgi_helper import MockDispatch

default_token = {
    "token_type": "Bearer",
    "access_token": "a",
    "refresh_token": "b",
    "expires_in": "3600",
    "expires_at": int(time.time()) + 3600,
}


def test_refresh_token():
    def verifier(request):
        content = request.form
        if str(request.url) == "https://provider.test/token":
            assert "assertion" in content

    with AssertionClient(
        "https://provider.test/token",
        issuer="foo",
        subject="foo",
        audience="foo",
        alg="HS256",
        key="secret",
        transport=WSGITransport(MockDispatch(default_token, assert_func=verifier)),
    ) as client:
        client.get("https://provider.test")

    # trigger more case
    now = int(time.time())
    with AssertionClient(
        "https://provider.test/token",
        issuer="foo",
        subject=None,
        audience="foo",
        issued_at=now,
        expires_at=now + 3600,
        header={"alg": "HS256"},
        key="secret",
        scope="email",
        claims={"test_mode": "true"},
        transport=WSGITransport(MockDispatch(default_token, assert_func=verifier)),
    ) as client:
        client.get("https://provider.test")
        client.get("https://provider.test")


def test_without_alg():
    with AssertionClient(
        "https://provider.test/token",
        issuer="foo",
        subject="foo",
        audience="foo",
        key="secret",
        transport=WSGITransport(MockDispatch(default_token)),
    ) as client:
        with pytest.raises(ValueError):
            client.get("https://provider.test")



import calendar
import datetime
import random
import re

from authlib.common.encoding import json_dumps
from authlib.common.encoding import json_loads
from authlib.common.encoding import to_bytes
from authlib.common.encoding import to_unicode

from ..errors import DecodeError
from ..errors import InsecureClaimError
from ..rfc7515 import JsonWebSignature
from ..rfc7516 import JsonWebEncryption
from ..rfc7517 import Key
from ..rfc7517 import KeySet
from .claims import JWTClaims


class JsonWebToken:
    SENSITIVE_NAMES = ("password", "token", "secret", "secret_key")
    # Thanks to sentry SensitiveDataFilter
    SENSITIVE_VALUES = re.compile(
        r"|".join(
            [
                # http://www.richardsramblings.com/regex/credit-card-numbers/
                r"\b(?:3[47]\d|(?:4\d|5[1-5]|65)\d{2}|6011)\d{12}\b",
                # various private keys
                r"-----BEGIN[A-Z ]+PRIVATE KEY-----.+-----END[A-Z ]+PRIVATE KEY-----",
                # social security numbers (US)
                r"^\b(?!(000|666|9))\d{3}-(?!00)\d{2}-(?!0000)\d{4}\b",
            ]
        ),
        re.DOTALL,
    )

    def __init__(self, algorithms, private_headers=None):
        self._jws = JsonWebSignature(algorithms, private_headers=private_headers)
        self._jwe = JsonWebEncryption(algorithms, private_headers=private_headers)

    def check_sensitive_data(self, payload):
        """Check if payload contains sensitive information."""
        for k in payload:
            # check claims key name
            if k in self.SENSITIVE_NAMES:
                raise InsecureClaimError(k)

            # check claims values
            v = payload[k]
            if isinstance(v, str) and self.SENSITIVE_VALUES.search(v):
                raise InsecureClaimError(k)

    def encode(self, header, payload, key, check=True):
        """Encode a JWT with the given header, payload and key.

        :param header: A dict of JWS header
        :param payload: A dict to be encoded
        :param key: key used to sign the signature
        :param check: check if sensitive data in payload
        :return: bytes
        """
        header.setdefault("typ", "JWT")

        for k in ["exp", "iat", "nbf"]:
            # convert datetime into timestamp
            claim = payload.get(k)
            if isinstance(claim, datetime.datetime):
                payload[k] = calendar.timegm(claim.utctimetuple())

        if check:
            self.check_sensitive_data(payload)

        key = find_encode_key(key, header)
        text = to_bytes(json_dumps(payload))
        if "enc" in header:
            return self._jwe.serialize_compact(header, text, key)
        else:
            return self._jws.serialize_compact(header, text, key)

    def decode(self, s, key, claims_cls=None, claims_options=None, claims_params=None):
        """Decode the JWT with the given key. This is similar with
        :meth:`verify`, except that it will raise BadSignatureError when
        signature doesn't match.

        :param s: text of JWT
        :param key: key used to verify the signature
        :param claims_cls: class to be used for JWT claims
        :param claims_options: `options` parameters for claims_cls
        :param claims_params: `params` parameters for claims_cls
        :return: claims_cls instance
        :raise: BadSignatureError
        """
        if claims_cls is None:
            claims_cls = JWTClaims

        if callable(key):
            load_key = key
        else:
            load_key = create_load_key(prepare_raw_key(key))

        s = to_bytes(s)
        dot_count = s.count(b".")
        if dot_count == 2:
            data = self._jws.deserialize_compact(s, load_key, decode_payload)
        elif dot_count == 4:
            data = self._jwe.deserialize_compact(s, load_key, decode_payload)
        else:
            raise DecodeError("Invalid input segments length")
        return claims_cls(
            data["payload"],
            data["header"],
            options=claims_options,
            params=claims_params,
        )


def decode_payload(bytes_payload):
    try:
        payload = json_loads(to_unicode(bytes_payload))
    except ValueError as exc:
        raise DecodeError("Invalid payload value") from exc
    if not isinstance(payload, dict):
        raise DecodeError("Invalid payload type")
    return payload


def prepare_raw_key(raw):
    if isinstance(raw, KeySet):
        return raw

    if isinstance(raw, str) and raw.startswith("{") and raw.endswith("}"):
        raw = json_loads(raw)
    elif isinstance(raw, (tuple, list)):
        raw = {"keys": raw}
    return raw


def find_encode_key(key, header):
    if isinstance(key, KeySet):
        kid = header.get("kid")
        if kid:
            return key.find_by_kid(kid)

        rv = random.choice(key.keys)
        # use side effect to add kid value into header
        header["kid"] = rv.kid
        return rv

    if isinstance(key, dict) and "keys" in key:
        keys = key["keys"]
        kid = header.get("kid")
        for k in keys:
            if k.get("kid") == kid:
                return k

        if not kid:
            rv = random.choice(keys)
            header["kid"] = rv["kid"]
            return rv
        raise ValueError("Invalid JSON Web Key Set")

    # append kid into header
    if isinstance(key, dict) and "kid" in key:
        header["kid"] = key["kid"]
    elif isinstance(key, Key) and key.kid:
        header["kid"] = key.kid
    return key


def create_load_key(key):
    def load_key(header, payload):
        if isinstance(key, KeySet):
            return key.find_by_kid(header.get("kid"))

        if isinstance(key, dict) and "keys" in key:
            keys = key["keys"]
            kid = header.get("kid")

            if kid is not None:
                # look for the requested key
                for k in keys:
                    if k.get("kid") == kid:
                        return k
            else:
                # use the only key
                if len(keys) == 1:
                    return keys[0]
            raise ValueError("Invalid JSON Web Key Set")
        return key

    return load_key



import pytest

from authlib.oauth1.rfc5849 import errors
from tests.util import decode_response

from .models import User


def test_invalid_authorization(factory, server):
    url = "/oauth/authorize"
    request = factory.post(url)
    with pytest.raises(errors.MissingRequiredParameterError):
        server.check_authorization_request(request)

    request = factory.post(url, data={"oauth_token": "a"})
    with pytest.raises(errors.InvalidTokenError):
        server.check_authorization_request(request)


def test_invalid_initiate(factory, server):
    url = "/oauth/initiate"
    # Test with non-existent client
    request = factory.post(
        url,
        data={
            "oauth_consumer_key": "nonexistent",  # Client doesn't exist
            "oauth_callback": "oob",
            "oauth_signature_method": "PLAINTEXT",
            "oauth_signature": "secret&",
        },
    )
    resp = server.create_temporary_credentials_response(request)
    data = decode_response(resp.content)
    assert data["error"] == "invalid_client"


def test_authorize_denied(factory, plaintext_server):
    server = plaintext_server
    initiate_url = "/oauth/initiate"
    authorize_url = "/oauth/authorize"

    # case 1
    request = factory.post(
        initiate_url,
        data={
            "oauth_consumer_key": "client",
            "oauth_callback": "oob",
            "oauth_signature_method": "PLAINTEXT",
            "oauth_signature": "secret&",
        },
    )
    resp = server.create_temporary_credentials_response(request)
    data = decode_response(resp.content)
    assert "oauth_token" in data

    request = factory.post(authorize_url, data={"oauth_token": data["oauth_token"]})
    resp = server.create_authorization_response(request)
    assert resp.status_code == 302
    assert "access_denied" in resp["Location"]
    assert "https://client.test" in resp["Location"]

    # case 2
    request = factory.post(
        initiate_url,
        data={
            "oauth_consumer_key": "client",
            "oauth_callback": "https://i.test",
            "oauth_signature_method": "PLAINTEXT",
            "oauth_signature": "secret&",
        },
    )
    resp = server.create_temporary_credentials_response(request)
    data = decode_response(resp.content)
    assert "oauth_token" in data
    request = factory.post(authorize_url, data={"oauth_token": data["oauth_token"]})
    resp = server.create_authorization_response(request)
    assert resp.status_code == 302
    assert "access_denied" in resp["Location"]
    assert "https://i.test" in resp["Location"]


def test_authorize_granted(factory, plaintext_server):
    server = plaintext_server
    user = User.objects.get(username="foo")
    initiate_url = "/oauth/initiate"
    authorize_url = "/oauth/authorize"

    # case 1
    request = factory.post(
        initiate_url,
        data={
            "oauth_consumer_key": "client",
            "oauth_callback": "oob",
            "oauth_signature_method": "PLAINTEXT",
            "oauth_signature": "secret&",
        },
    )
    resp = server.create_temporary_credentials_response(request)
    data = decode_response(resp.content)
    assert "oauth_token" in data

    request = factory.post(authorize_url, data={"oauth_token": data["oauth_token"]})
    resp = server.create_authorization_response(request, user)
    assert resp.status_code == 302

    assert "oauth_verifier" in resp["Location"]
    assert "https://client.test" in resp["Location"]

    # case 2
    request = factory.post(
        initiate_url,
        data={
            "oauth_consumer_key": "client",
            "oauth_callback": "https://i.test",
            "oauth_signature_method": "PLAINTEXT",
            "oauth_signature": "secret&",
        },
    )
    resp = server.create_temporary_credentials_response(request)
    data = decode_response(resp.content)
    assert "oauth_token" in data

    request = factory.post(authorize_url, data={"oauth_token": data["oauth_token"]})
    resp = server.create_authorization_response(request, user)

    assert resp.status_code == 302
    assert "oauth_verifier" in resp["Location"]
    assert "https://i.test" in resp["Location"]



import logging

from authlib.common.urls import add_params_to_uri

from ..errors import AccessDeniedError
from ..errors import OAuth2Error
from ..errors import UnauthorizedClientError
from ..hooks import hooked
from .base import AuthorizationEndpointMixin
from .base import BaseGrant

log = logging.getLogger(__name__)


class ImplicitGrant(BaseGrant, AuthorizationEndpointMixin):
    """The implicit grant type is used to obtain access tokens (it does not
    support the issuance of refresh tokens) and is optimized for public
    clients known to operate a particular redirection URI.  These clients
    are typically implemented in a browser using a scripting language
    such as JavaScript.

    Since this is a redirection-based flow, the client must be capable of
    interacting with the resource owner's user-agent (typically a web
    browser) and capable of receiving incoming requests (via redirection)
    from the authorization server.

    Unlike the authorization code grant type, in which the client makes
    separate requests for authorization and for an access token, the
    client receives the access token as the result of the authorization
    request.

    The implicit grant type does not include client authentication, and
    relies on the presence of the resource owner and the registration of
    the redirection URI.  Because the access token is encoded into the
    redirection URI, it may be exposed to the resource owner and other
    applications residing on the same device::

        +----------+
        | Resource |
        |  Owner   |
        |          |
        +----------+
             ^
             |
            (B)
        +----|-----+          Client Identifier     +---------------+
        |         -+----(A)-- & Redirection URI --->|               |
        |  User-   |                                | Authorization |
        |  Agent  -|----(B)-- User authenticates -->|     Server    |
        |          |                                |               |
        |          |<---(C)--- Redirection URI ----<|               |
        |          |          with Access Token     +---------------+
        |          |            in Fragment
        |          |                                +---------------+
        |          |----(D)--- Redirection URI ---->|   Web-Hosted  |
        |          |          without Fragment      |     Client    |
        |          |                                |    Resource   |
        |     (F)  |<---(E)------- Script ---------<|               |
        |          |                                +---------------+
        +-|--------+
          |    |
         (A)  (G) Access Token
          |    |
          ^    v
        +---------+
        |         |
        |  Client |
        |         |
        +---------+
    """

    #: authorization_code grant type has authorization endpoint
    AUTHORIZATION_ENDPOINT = True
    #: Allowed client auth methods for token endpoint
    TOKEN_ENDPOINT_AUTH_METHODS = ["none"]

    RESPONSE_TYPES = {"token"}
    GRANT_TYPE = "implicit"
    ERROR_RESPONSE_FRAGMENT = True

    @hooked
    def validate_authorization_request(self):
        """The client constructs the request URI by adding the following
        parameters to the query component of the authorization endpoint URI
        using the "application/x-www-form-urlencoded" format.
        Per `Section 4.2.1`_.

        response_type
             REQUIRED.  Value MUST be set to "token".

        client_id
             REQUIRED.  The client identifier as described in Section 2.2.

        redirect_uri
             OPTIONAL.  As described in Section 3.1.2.

        scope
             OPTIONAL.  The scope of the access request as described by
             Section 3.3.

        state
             RECOMMENDED.  An opaque value used by the client to maintain
             state between the request and callback.  The authorization
             server includes this value when redirecting the user-agent back
             to the client.  The parameter SHOULD be used for preventing
             cross-site request forgery as described in Section 10.12.

        The client directs the resource owner to the constructed URI using an
        HTTP redirection response, or by other means available to it via the
        user-agent.

        For example, the client directs the user-agent to make the following
        HTTP request using TLS:

        .. code-block:: http

            GET /authorize?response_type=token&client_id=s6BhdRkqt3&state=xyz
            &redirect_uri=https%3A%2F%2Fclient%2Eexample%2Ecom%2Fcb HTTP/1.1
            Host: server.example.com

        .. _`Section 4.2.1`: https://tools.ietf.org/html/rfc6749#section-4.2.1
        """
        # ignore validate for response_type, since it is validated by
        # check_authorization_endpoint

        # The implicit grant type is optimized for public clients
        client = self.authenticate_token_endpoint_client()
        log.debug("Validate authorization request of %r", client)

        redirect_uri = self.validate_authorization_redirect_uri(self.request, client)

        response_type = self.request.payload.response_type
        if not client.check_response_type(response_type):
            raise UnauthorizedClientError(
                f"The client is not authorized to use 'response_type={response_type}'",
                redirect_uri=redirect_uri,
                redirect_fragment=True,
            )

        try:
            self.request.client = client
            self.validate_requested_scope()
        except OAuth2Error as error:
            error.redirect_uri = redirect_uri
            error.redirect_fragment = True
            raise error
        return redirect_uri

    @hooked
    def create_authorization_response(self, redirect_uri, grant_user):
        """If the resource owner grants the access request, the authorization
        server issues an access token and delivers it to the client by adding
        the following parameters to the fragment component of the redirection
        URI using the "application/x-www-form-urlencoded" format.
        Per `Section 4.2.2`_.

        access_token
             REQUIRED.  The access token issued by the authorization server.

        token_type
             REQUIRED.  The type of the token issued as described in
             Section 7.1.  Value is case insensitive.

        expires_in
             RECOMMENDED.  The lifetime in seconds of the access token.  For
             example, the value "3600" denotes that the access token will
             expire in one hour from the time the response was generated.
             If omitted, the authorization server SHOULD provide the
             expiration time via other means or document the default value.

        scope
             OPTIONAL, if identical to the scope requested by the client;
             otherwise, REQUIRED.  The scope of the access token as
             described by Section 3.3.

        state
             REQUIRED if the "state" parameter was present in the client
             authorization request.  The exact value received from the
             client.

        The authorization server MUST NOT issue a refresh token.

        For example, the authorization server redirects the user-agent by
        sending the following HTTP response:

        .. code-block:: http

            HTTP/1.1 302 Found
            Location: http://example.com/cb#access_token=2YotnFZFEjr1zCsicMWpAA
                   &state=xyz&token_type=example&expires_in=3600

        Developers should note that some user-agents do not support the
        inclusion of a fragment component in the HTTP "Location" response
        header field.  Such clients will require using other methods for
        redirecting the client than a 3xx redirection response -- for
        example, returning an HTML page that includes a 'continue' button
        with an action linked to the redirection URI.

        .. _`Section 4.2.2`: https://tools.ietf.org/html/rfc6749#section-4.2.2

        :param redirect_uri: Redirect to the given URI for the authorization
        :param grant_user: if resource owner granted the request, pass this
            resource owner, otherwise pass None.
        :returns: (status_code, body, headers)
        """
        state = self.request.payload.state
        if grant_user:
            self.request.user = grant_user
            token = self.generate_token(
                user=grant_user,
                scope=self.request.payload.scope,
                include_refresh_token=False,
            )
            log.debug("Grant token %r to %r", token, self.request.client)

            self.save_token(token)
            params = [(k, token[k]) for k in token]
            if state:
                params.append(("state", state))

            uri = add_params_to_uri(redirect_uri, params, fragment=True)
            headers = [("Location", uri)]
            return 302, "", headers
        else:
            raise AccessDeniedError(redirect_uri=redirect_uri, redirect_fragment=True)



import base64
import binascii
from urllib.parse import unquote

from authlib.common.encoding import to_unicode


def list_to_scope(scope):
    """Convert a list of scopes to a space separated string."""
    if isinstance(scope, (set, tuple, list)):
        return " ".join([to_unicode(s) for s in scope])
    if scope is None:
        return scope
    return to_unicode(scope)


def scope_to_list(scope):
    """Convert a space separated string to a list of scopes."""
    if isinstance(scope, (tuple, list, set)):
        return [to_unicode(s) for s in scope]
    elif scope is None:
        return None
    return scope.strip().split()


def extract_basic_authorization(headers):
    auth = headers.get("Authorization")
    if not auth or " " not in auth:
        return None, None

    auth_type, auth_token = auth.split(None, 1)
    if auth_type.lower() != "basic":
        return None, None

    try:
        query = to_unicode(base64.b64decode(auth_token))
    except (binascii.Error, TypeError):
        return None, None
    if ":" in query:
        username, password = query.split(":", 1)
        return unquote(username), unquote(password)
    return query, None



import time
from unittest import mock

from authlib.jose import jwt
from authlib.oauth2.rfc7523 import PrivateKeyJWT
from tests.util import read_file_path

public_key = read_file_path("rsa_public.pem")
private_key = read_file_path("rsa_private.pem")


def test_nothing_set():
    jwt_signer = PrivateKeyJWT()

    assert jwt_signer.token_endpoint is None
    assert jwt_signer.claims is None
    assert jwt_signer.headers is None
    assert jwt_signer.alg == "RS256"


def test_endpoint_set():
    jwt_signer = PrivateKeyJWT(
        token_endpoint="https://provider.test/oauth/access_token"
    )

    assert jwt_signer.token_endpoint == "https://provider.test/oauth/access_token"
    assert jwt_signer.claims is None
    assert jwt_signer.headers is None
    assert jwt_signer.alg == "RS256"


def test_alg_set():
    jwt_signer = PrivateKeyJWT(alg="RS512")

    assert jwt_signer.token_endpoint is None
    assert jwt_signer.claims is None
    assert jwt_signer.headers is None
    assert jwt_signer.alg == "RS512"


def test_claims_set():
    jwt_signer = PrivateKeyJWT(claims={"foo1": "bar1"})

    assert jwt_signer.token_endpoint is None
    assert jwt_signer.claims == {"foo1": "bar1"}
    assert jwt_signer.headers is None
    assert jwt_signer.alg == "RS256"


def test_headers_set():
    jwt_signer = PrivateKeyJWT(headers={"foo1": "bar1"})

    assert jwt_signer.token_endpoint is None
    assert jwt_signer.claims is None
    assert jwt_signer.headers == {"foo1": "bar1"}
    assert jwt_signer.alg == "RS256"


def test_all_set():
    jwt_signer = PrivateKeyJWT(
        token_endpoint="https://provider.test/oauth/access_token",
        claims={"foo1a": "bar1a"},
        headers={"foo1b": "bar1b"},
        alg="RS512",
    )

    assert jwt_signer.token_endpoint == "https://provider.test/oauth/access_token"
    assert jwt_signer.claims == {"foo1a": "bar1a"}
    assert jwt_signer.headers == {"foo1b": "bar1b"}
    assert jwt_signer.alg == "RS512"


def sign_and_decode(jwt_signer, client_id, public_key, private_key, token_endpoint):
    auth = mock.MagicMock()
    auth.client_id = client_id
    auth.client_secret = private_key

    pre_sign_time = int(time.time())

    data = jwt_signer.sign(auth, token_endpoint).decode("utf-8")
    decoded = jwt.decode(
        data, public_key
    )  # , claims_cls=None, claims_options=None, claims_params=None):

    iat = decoded.pop("iat")
    exp = decoded.pop("exp")
    jti = decoded.pop("jti")

    return decoded, pre_sign_time, iat, exp, jti


def test_sign_nothing_set():
    jwt_signer = PrivateKeyJWT()

    decoded, pre_sign_time, iat, exp, jti = sign_and_decode(
        jwt_signer,
        "client_id_1",
        public_key,
        private_key,
        "https://provider.test/oauth/access_token",
    )

    assert iat >= pre_sign_time
    assert exp >= iat + 3600
    assert exp <= iat + 3600 + 2
    assert jti is not None

    assert {
        "iss": "client_id_1",
        "aud": "https://provider.test/oauth/access_token",
        "sub": "client_id_1",
    } == decoded
    assert {"alg": "RS256", "typ": "JWT"} == decoded.header


def test_sign_custom_jti():
    jwt_signer = PrivateKeyJWT(claims={"jti": "custom_jti"})

    decoded, pre_sign_time, iat, exp, jti = sign_and_decode(
        jwt_signer,
        "client_id_1",
        public_key,
        private_key,
        "https://provider.test/oauth/access_token",
    )

    assert iat >= pre_sign_time
    assert exp >= iat + 3600
    assert exp <= iat + 3600 + 2
    assert "custom_jti" == jti

    assert decoded == {
        "iss": "client_id_1",
        "aud": "https://provider.test/oauth/access_token",
        "sub": "client_id_1",
    }
    assert {"alg": "RS256", "typ": "JWT"} == decoded.header


def test_sign_with_additional_header():
    jwt_signer = PrivateKeyJWT(headers={"kid": "custom_kid"})

    decoded, pre_sign_time, iat, exp, jti = sign_and_decode(
        jwt_signer,
        "client_id_1",
        public_key,
        private_key,
        "https://provider.test/oauth/access_token",
    )

    assert iat >= pre_sign_time
    assert exp >= iat + 3600
    assert exp <= iat + 3600 + 2
    assert jti is not None

    assert decoded == {
        "iss": "client_id_1",
        "aud": "https://provider.test/oauth/access_token",
        "sub": "client_id_1",
    }
    assert {"alg": "RS256", "typ": "JWT", "kid": "custom_kid"} == decoded.header


def test_sign_with_additional_headers():
    jwt_signer = PrivateKeyJWT(
        headers={"kid": "custom_kid", "jku": "https://provider.test/oauth/jwks"}
    )

    decoded, pre_sign_time, iat, exp, jti = sign_and_decode(
        jwt_signer,
        "client_id_1",
        public_key,
        private_key,
        "https://provider.test/oauth/access_token",
    )

    assert iat >= pre_sign_time
    assert exp >= iat + 3600
    assert exp <= iat + 3600 + 2
    assert jti is not None

    assert decoded == {
        "iss": "client_id_1",
        "aud": "https://provider.test/oauth/access_token",
        "sub": "client_id_1",
    }
    assert {
        "alg": "RS256",
        "typ": "JWT",
        "kid": "custom_kid",
        "jku": "https://provider.test/oauth/jwks",
    } == decoded.header


def test_sign_with_additional_claim():
    jwt_signer = PrivateKeyJWT(claims={"name": "Foo"})

    decoded, pre_sign_time, iat, exp, jti = sign_and_decode(
        jwt_signer,
        "client_id_1",
        public_key,
        private_key,
        "https://provider.test/oauth/access_token",
    )

    assert iat >= pre_sign_time
    assert exp >= iat + 3600
    assert exp <= iat + 3600 + 2
    assert jti is not None

    assert decoded == {
        "iss": "client_id_1",
        "aud": "https://provider.test/oauth/access_token",
        "sub": "client_id_1",
        "name": "Foo",
    }
    assert {"alg": "RS256", "typ": "JWT"} == decoded.header


def test_sign_with_additional_claims():
    jwt_signer = PrivateKeyJWT(claims={"name": "Foo", "role": "bar"})

    decoded, pre_sign_time, iat, exp, jti = sign_and_decode(
        jwt_signer,
        "client_id_1",
        public_key,
        private_key,
        "https://provider.test/oauth/access_token",
    )

    assert iat >= pre_sign_time
    assert exp >= iat + 3600
    assert exp <= iat + 3600 + 2
    assert jti is not None

    assert decoded == {
        "iss": "client_id_1",
        "aud": "https://provider.test/oauth/access_token",
        "sub": "client_id_1",
        "name": "Foo",
        "role": "bar",
    }
    assert {"alg": "RS256", "typ": "JWT"} == decoded.header
